from collections import deque
from datetime import datetime, timedelta
from time import perf_counter
from typing import ClassVar, NamedTuple


class DurationRecord(NamedTuple):
    duration: float
    written_at: float


class DurationRecorder:
    task_durations: ClassVar[dict[str, deque[DurationRecord]]] = {}

    def __init__(self, task_name: str):
        self._task_name = task_name
        self._start_time = perf_counter()
        self._did_record = False

    def record(self):
        # Check that this is the first time.
        if self._did_record:
            raise RuntimeError("Cannot record more than once")
        self._did_record = True

        # Get the task name and current time.
        task_name = self._task_name
        now_time = perf_counter()

        # Get the deque.
        record_deque = self.task_durations.get(task_name)
        if record_deque is None:
            record_deque = deque[DurationRecord](maxlen=1024)
            self.task_durations[task_name] = record_deque

        # Add the record.
        duration_record = DurationRecord(
            duration=now_time - self._start_time,
            written_at=now_time,
        )
        record_deque.append(duration_record)

        # Remove items that are more than a minute old.
        preserve_from = now_time - 60.0
        while record_deque and record_deque[0][1] < preserve_from:
            record_deque.popleft()


def to_moment(exact_time: datetime) -> datetime:
    """
    In Solie's terminlogy, moment refers to a time on a 10-second unit.
    This is because one candlestick holds 10 seconds of information.
    """
    moment = exact_time.replace(microsecond=0)
    moment = moment - timedelta(seconds=moment.second % 10)
    return moment



import math
import random
from datetime import datetime, timedelta, timezone
from enum import Enum
from itertools import product
from multiprocessing.managers import ListProxy
from typing import Any, NamedTuple

import numpy as np
import pandas as pd

from .data_models import (
    AccountState,
    Decision,
    DecisionInput,
    IndicatorInput,
    OpenOrder,
    OrderType,
    Position,
    PositionDirection,
    Strategy,
    VirtualPlacement,
    VirtualState,
)
from .user_settings import SavedStrategy

GRAPH_TYPES = ["PRICE", "VOLUME", "ABSTRACT"]


def make_indicators(
    strategy: Strategy,
    target_symbols: list[str],
    candle_data: pd.DataFrame,
    only_last_index: bool = False,
) -> pd.DataFrame:
    # ■■■■■ interpolate nans ■■■■■

    candle_data = candle_data.interpolate()

    # ■■■■■ make dummy row to avoid ta error with all nan series ■■■■■

    if len(candle_data) > 0:
        dummy_index = candle_data.index[-1] + timedelta(seconds=1)
    else:
        dummy_index = datetime.fromtimestamp(0, tz=timezone.utc)

    candle_data.loc[dummy_index, :] = 0.0

    # ■■■■■ basic values ■■■■■

    blank_column_trios = product(
        target_symbols,
        ("PRICE", "VOLUME", "ABSTRACT"),
        ("BLANK",),
    )
    new_indicators: dict[str, pd.Series] = {}
    base_index = candle_data.index
    for blank_column in ("/".join(t) for t in blank_column_trios):
        new_indicators[blank_column] = pd.Series(
            np.nan,
            index=base_index,
            dtype=np.float32,
        )

    # ■■■■■ make individual indicators ■■■■■

    input = IndicatorInput(
        target_symbols=target_symbols,
        candle_data=candle_data,
        new_indicators=new_indicators,
    )
    strategy.create_indicators(input)

    # ■■■■■ concatenate individual indicators into one ■■■■■

    for column_name, new_indicator in new_indicators.items():
        # Validate the name format.
        if not isinstance(column_name, str):
            continue
        column_trio = column_name.split("/")
        if len(column_trio) != 3:
            continue
        symbol, category, _ = column_trio
        if symbol not in target_symbols or category not in GRAPH_TYPES:
            continue
        # Validate the indicator format.
        if not isinstance(new_indicator, pd.Series):
            continue
        if not pd.api.types.is_numeric_dtype(new_indicator):
            continue
        # Convert each element into strings and make it into a name.
        new_indicator.name = column_name

    indicators = pd.concat(new_indicators.values(), axis="columns")
    indicators = indicators.astype(np.float32)

    # ■■■■■ remove dummy row ■■■■■

    indicators = indicators.iloc[:-1]

    if only_last_index:
        return indicators.tail(1)
    else:
        return indicators


def make_decisions(
    strategy: Strategy,
    target_symbols: list[str],
    current_moment: datetime,
    current_candle_data: dict[str, float],
    current_indicators: dict[str, float],
    account_state: AccountState,
    scribbles: dict[Any, Any],
) -> dict[str, dict[OrderType, Decision]]:
    # ■■■■■ decision template ■■■■■

    new_decisions: dict[str, dict[OrderType, Decision]] = {}
    for symbol in target_symbols:
        new_decisions[symbol] = {}

    # ■■■■■ write decisions ■■■■■

    input = DecisionInput(
        target_symbols=target_symbols,
        account_state=account_state,
        current_moment=current_moment,
        current_candle_data=current_candle_data,
        current_indicators=current_indicators,
        scribbles=scribbles,
        new_decisions=new_decisions,
    )
    strategy.create_decisions(input)

    # ■■■■■ return decision ■■■■■

    blank_symbols: list[str] = []
    for symbol, symbol_decisions in new_decisions.items():
        if len(symbol_decisions) == 0:
            blank_symbols.append(symbol)
    for blank_symbol in blank_symbols:
        new_decisions.pop(blank_symbol)

    return new_decisions


class SimulationError(Exception):
    pass


class OrderRole(Enum):
    MAKER = "MAKER"
    TAKER = "TAKER"


class CalculationInput(NamedTuple):
    strategy: Strategy
    progress_list: ListProxy
    target_progress: int
    target_symbols: list[str]
    calculation_index: pd.DatetimeIndex
    chunk_candle_data: pd.DataFrame
    chunk_indicators: pd.DataFrame
    chunk_asset_record: pd.DataFrame
    chunk_unrealized_changes: pd.Series
    chunk_scribbles: dict[Any, Any]
    chunk_account_state: AccountState
    chunk_virtual_state: VirtualState


class CalculationOutput(NamedTuple):
    chunk_asset_record: pd.DataFrame
    chunk_unrealized_changes: pd.Series
    chunk_scribbles: dict[Any, Any]
    chunk_account_state: AccountState
    chunk_virtual_state: VirtualState


ORDER_ID_MIN, ORDER_ID_MAX = 10**18, 10**19 - 1


def simulate_chunk(calculation_input: CalculationInput) -> CalculationOutput:
    # leverage is treated as 1
    # because those are going to be applied at the presentation phase

    # ■■■■■ get data ■■■■■

    strategy = calculation_input.strategy
    progress_list = calculation_input.progress_list
    target_progress = calculation_input.target_progress
    target_symbols = calculation_input.target_symbols
    calculation_index = calculation_input.calculation_index
    chunk_candle_data = calculation_input.chunk_candle_data
    chunk_indicators = calculation_input.chunk_indicators
    chunk_asset_record = calculation_input.chunk_asset_record
    chunk_unrealized_changes = calculation_input.chunk_unrealized_changes
    chunk_scribbles = calculation_input.chunk_scribbles
    chunk_account_state = calculation_input.chunk_account_state
    chunk_virtual_state = calculation_input.chunk_virtual_state

    if isinstance(strategy, SavedStrategy):
        strategy.compile_code()

    # ■■■■■ basic values ■■■■■

    decision_lag = 3000  # milliseconds

    # ■■■■■ return blank data if there's nothing to calculate ■■■■■

    if len(calculation_index) == 0:
        calculation_output = CalculationOutput(
            chunk_asset_record=chunk_asset_record,
            chunk_unrealized_changes=chunk_unrealized_changes,
            chunk_scribbles=chunk_scribbles,
            chunk_account_state=chunk_account_state,
            chunk_virtual_state=chunk_virtual_state,
        )
        return calculation_output

    # ■■■■■ convert to numpy objects for fast calculation ■■■■■

    calculation_index_ar = calculation_index.to_numpy()  # inside are datetime objects
    candle_data_ar = chunk_candle_data.to_records()
    indicators_ar = chunk_indicators.to_records()
    asset_record_ar = chunk_asset_record.to_records()
    chunk_unrealized_changes_ar = chunk_unrealized_changes.to_frame().to_records()

    # ■■■■■ actual loop calculation ■■■■■

    calculation_index_length = len(calculation_index_ar)
    first_calculation_moment = calculation_index_ar[0]

    for cycle in range(calculation_index_length):
        before_moment = calculation_index_ar[cycle]
        current_moment = before_moment + timedelta(seconds=10)

        for symbol in target_symbols:
            # ■■■■■ basic variables ■■■■■

            open_price = candle_data_ar[cycle][f"{symbol}/OPEN"]
            close_price = candle_data_ar[cycle][f"{symbol}/CLOSE"]
            if math.isnan(open_price) or math.isnan(close_price):
                continue

            would_trade_happen = False
            is_new_trade_found = False
            amount_shift = 0
            fill_price = 0
            role: OrderRole | None = None

            # ■■■■■ check if any order would be filled ■■■■■

            price_speed = (close_price - open_price) / 10
            is_margin_negative = False
            is_margin_nan = False

            # special placements
            if OrderType.CANCEL_ALL in chunk_virtual_state.placements[symbol]:
                cancel_placement_names: list[OrderType] = []
                for order_type in chunk_virtual_state.placements[symbol].keys():
                    if order_type.is_later() or order_type.is_book():
                        cancel_placement_names.append(order_type)
                for cancel_placement_name in cancel_placement_names:
                    chunk_virtual_state.placements[symbol].pop(cancel_placement_name)
                chunk_virtual_state.placements[symbol].pop(OrderType.CANCEL_ALL)

            # instant placements
            if OrderType.NOW_CLOSE in chunk_virtual_state.placements[symbol]:
                would_trade_happen = True
                placement = chunk_virtual_state.placements[symbol][OrderType.NOW_CLOSE]
                role = OrderRole.TAKER
                fill_price = open_price + price_speed * (decision_lag / 1000)
                amount_shift = -chunk_virtual_state.positions[symbol].amount
                chunk_virtual_state.placements[symbol].pop(OrderType.NOW_CLOSE)

            if OrderType.NOW_BUY in chunk_virtual_state.placements[symbol]:
                would_trade_happen = True
                placement = chunk_virtual_state.placements[symbol][OrderType.NOW_BUY]
                role = OrderRole.TAKER
                fill_price = open_price + price_speed * (decision_lag / 1000)
                fill_margin = placement.margin
                if fill_margin < 0:
                    is_margin_negative = True
                if math.isnan(fill_margin):
                    is_margin_nan = True
                amount_shift = fill_margin / fill_price
                chunk_virtual_state.placements[symbol].pop(OrderType.NOW_BUY)

            if OrderType.NOW_SELL in chunk_virtual_state.placements[symbol]:
                would_trade_happen = True
                placement = chunk_virtual_state.placements[symbol][OrderType.NOW_SELL]
                role = OrderRole.TAKER
                fill_price = open_price + price_speed * (decision_lag / 1000)
                fill_margin = placement.margin
                if fill_margin < 0:
                    is_margin_negative = True
                if math.isnan(fill_margin):
                    is_margin_nan = True
                amount_shift = -fill_margin / fill_price
                chunk_virtual_state.placements[symbol].pop(OrderType.NOW_SELL)

            # conditional placements
            if OrderType.LATER_UP_CLOSE in chunk_virtual_state.placements[symbol]:
                placement = chunk_virtual_state.placements[symbol][
                    OrderType.LATER_UP_CLOSE
                ]
                boundary = placement.boundary

                wobble_high = candle_data_ar[cycle][f"{symbol}/HIGH"]
                wobble_low = candle_data_ar[cycle][f"{symbol}/LOW"]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = OrderRole.TAKER
                    fill_price = boundary
                    amount_shift = -chunk_virtual_state.positions[symbol].amount
                    chunk_virtual_state.placements[symbol].pop(OrderType.LATER_UP_CLOSE)

            if OrderType.LATER_DOWN_CLOSE in chunk_virtual_state.placements[symbol]:
                placement = chunk_virtual_state.placements[symbol][
                    OrderType.LATER_DOWN_CLOSE
                ]
                boundary = placement.boundary

                wobble_high = candle_data_ar[cycle][f"{symbol}/HIGH"]
                wobble_low = candle_data_ar[cycle][f"{symbol}/LOW"]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = OrderRole.TAKER
                    fill_price = boundary
                    amount_shift = -chunk_virtual_state.positions[symbol].amount
                    chunk_virtual_state.placements[symbol].pop(
                        OrderType.LATER_DOWN_CLOSE
                    )

            if OrderType.LATER_UP_BUY in chunk_virtual_state.placements[symbol]:
                placement = chunk_virtual_state.placements[symbol][
                    OrderType.LATER_UP_BUY
                ]
                boundary = placement.boundary

                wobble_high = candle_data_ar[cycle][f"{symbol}/HIGH"]
                wobble_low = candle_data_ar[cycle][f"{symbol}/LOW"]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = OrderRole.TAKER
                    fill_price = boundary
                    fill_margin = placement.margin
                    if fill_margin < 0:
                        is_margin_negative = True
                    if math.isnan(fill_margin):
                        is_margin_nan = True
                    amount_shift = fill_margin / fill_price
                    chunk_virtual_state.placements[symbol].pop(OrderType.LATER_UP_BUY)

            if OrderType.LATER_DOWN_BUY in chunk_virtual_state.placements[symbol]:
                placement = chunk_virtual_state.placements[symbol][
                    OrderType.LATER_DOWN_BUY
                ]
                boundary = placement.boundary

                wobble_high = candle_data_ar[cycle][f"{symbol}/HIGH"]
                wobble_low = candle_data_ar[cycle][f"{symbol}/LOW"]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = OrderRole.TAKER
                    fill_price = boundary
                    fill_margin = placement.margin
                    if fill_margin < 0:
                        is_margin_negative = True
                    if math.isnan(fill_margin):
                        is_margin_nan = True
                    amount_shift = fill_margin / fill_price
                    chunk_virtual_state.placements[symbol].pop(OrderType.LATER_DOWN_BUY)

            if OrderType.LATER_UP_SELL in chunk_virtual_state.placements[symbol]:
                placement = chunk_virtual_state.placements[symbol][
                    OrderType.LATER_UP_SELL
                ]
                boundary = placement.boundary

                wobble_high = candle_data_ar[cycle][f"{symbol}/HIGH"]
                wobble_low = candle_data_ar[cycle][f"{symbol}/LOW"]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = OrderRole.TAKER
                    fill_price = boundary
                    fill_margin = placement.margin
                    if fill_margin < 0:
                        is_margin_negative = True
                    if math.isnan(fill_margin):
                        is_margin_nan = True
                    amount_shift = -fill_margin / fill_price
                    chunk_virtual_state.placements[symbol].pop(OrderType.LATER_UP_SELL)

            if OrderType.LATER_DOWN_SELL in chunk_virtual_state.placements[symbol]:
                placement = chunk_virtual_state.placements[symbol][
                    OrderType.LATER_DOWN_SELL
                ]
                boundary = placement.boundary

                wobble_high = candle_data_ar[cycle][f"{symbol}/HIGH"]
                wobble_low = candle_data_ar[cycle][f"{symbol}/LOW"]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = OrderRole.TAKER
                    fill_price = boundary
                    fill_margin = placement.margin
                    if fill_margin < 0:
                        is_margin_negative = True
                    if math.isnan(fill_margin):
                        is_margin_nan = True
                    amount_shift = -fill_margin / fill_price
                    chunk_virtual_state.placements[symbol].pop(
                        OrderType.LATER_DOWN_SELL
                    )

            if OrderType.BOOK_BUY in chunk_virtual_state.placements[symbol]:
                placement = chunk_virtual_state.placements[symbol][OrderType.BOOK_BUY]
                boundary = placement.boundary

                wobble_high = candle_data_ar[cycle][f"{symbol}/HIGH"]
                wobble_low = candle_data_ar[cycle][f"{symbol}/LOW"]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = OrderRole.MAKER
                    fill_price = boundary
                    fill_margin = placement.margin
                    if fill_margin < 0:
                        is_margin_negative = True
                    if math.isnan(fill_margin):
                        is_margin_nan = True
                    amount_shift = fill_margin / fill_price
                    chunk_virtual_state.placements[symbol].pop(OrderType.BOOK_BUY)

            if OrderType.BOOK_SELL in chunk_virtual_state.placements[symbol]:
                placement = chunk_virtual_state.placements[symbol][OrderType.BOOK_SELL]
                boundary = placement.boundary

                wobble_high = candle_data_ar[cycle][f"{symbol}/HIGH"]
                wobble_low = candle_data_ar[cycle][f"{symbol}/LOW"]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = OrderRole.MAKER
                    fill_price = boundary
                    fill_margin = placement.margin
                    if fill_margin < 0:
                        is_margin_negative = True
                    if math.isnan(fill_margin):
                        is_margin_nan = True
                    amount_shift = -fill_margin / fill_price
                    chunk_virtual_state.placements[symbol].pop(OrderType.BOOK_SELL)

            # check if situation is okay
            if is_margin_negative:
                text = ""
                text += "Got an order with a negative margin"
                text += f" while calculating {symbol} market at {current_moment}"
                raise SimulationError(text)

            elif is_margin_nan:
                text = ""
                text += "Got an order with a non-numeric margin"
                text += f" while calculating {symbol} market at {current_moment}"
                raise SimulationError(text)

            # ■■■■■ mimic the real world phenomenon ■■■■■

            if would_trade_happen:
                virtual_position = chunk_virtual_state.positions[symbol]
                before_entry_price = virtual_position.entry_price
                before_amount = virtual_position.amount

                virtual_position.amount += amount_shift
                current_amount = virtual_position.amount

                # case when the position is created from 0
                if before_amount == 0 and current_amount != 0:
                    virtual_position.entry_price = fill_price
                    invested_margin = abs(current_amount) * fill_price
                    chunk_virtual_state.available_balance -= invested_margin
                # case when the position is closed from something
                elif before_amount != 0 and current_amount == 0:
                    virtual_position.entry_price = 0
                    price_difference = fill_price - before_entry_price
                    realized_profit = price_difference * before_amount
                    returned_margin = abs(before_amount) * before_entry_price
                    chunk_virtual_state.available_balance += returned_margin
                    chunk_virtual_state.available_balance += realized_profit
                # case when the position direction is flipped
                elif before_amount * current_amount < 0:
                    virtual_position.entry_price = fill_price
                    price_difference = fill_price - before_entry_price
                    realized_profit = price_difference * before_amount
                    returned_margin = abs(before_amount) * before_entry_price
                    invested_margin = abs(current_amount) * fill_price
                    chunk_virtual_state.available_balance += returned_margin
                    chunk_virtual_state.available_balance -= invested_margin
                    chunk_virtual_state.available_balance += realized_profit
                # case when the position size is increased one the same direction
                elif abs(current_amount) > abs(before_amount):
                    before_numerator = before_entry_price * before_amount
                    new_numerator = fill_price * amount_shift
                    current_numerator = before_numerator + new_numerator
                    new_entry_price = current_numerator / current_amount
                    virtual_position.entry_price = new_entry_price
                    realized_profit = 0
                    invested_margin = abs(amount_shift) * fill_price
                    chunk_virtual_state.available_balance -= invested_margin
                    chunk_virtual_state.available_balance += realized_profit
                # case when the position size is decreased one the same direction
                else:
                    virtual_position.entry_price = before_entry_price
                    price_difference = fill_price - before_entry_price
                    realized_profit = price_difference * (-amount_shift)
                    returned_margin = abs(amount_shift) * before_entry_price
                    chunk_virtual_state.available_balance += returned_margin
                    chunk_virtual_state.available_balance += realized_profit

                is_new_trade_found = True

                if chunk_virtual_state.available_balance < 0:
                    text = ""
                    text += "Available balance went below zero"
                    text += f" while calculating {symbol} market"
                    text += f" at {current_moment}"
                    raise SimulationError(text)

            # ■■■■■ update the account state (symbol dependent) ■■■■■

            # locations
            current_entry_price = chunk_virtual_state.positions[symbol].entry_price
            current_entry_price = current_entry_price
            current_amount = chunk_virtual_state.positions[symbol].amount
            current_margin = abs(current_amount) * current_entry_price
            current_margin = current_margin
            if current_amount > 0.0:
                current_direction = PositionDirection.LONG
            elif current_amount < 0.0:
                current_direction = PositionDirection.SHORT
            else:
                current_direction = PositionDirection.NONE
            symbol_position = Position(
                margin=current_margin,
                direction=current_direction,
                entry_price=current_entry_price,
                update_time=datetime.fromtimestamp(0),
            )
            chunk_account_state.positions[symbol] = symbol_position

            # placements
            symbol_placements = chunk_virtual_state.placements[symbol]
            symbol_open_orders: dict[int, OpenOrder] = {}
            for order_type, placement in symbol_placements.items():
                order_id = placement.order_id
                boundary = placement.boundary
                left_margin = placement.margin
                symbol_open_orders[order_id] = OpenOrder(
                    order_type=order_type,
                    boundary=boundary,
                    left_margin=left_margin,
                )
            chunk_account_state.open_orders[symbol] = symbol_open_orders

            # ■■■■■ record (symbol dependent) ■■■■■

            if is_new_trade_found:
                fill_time = before_moment + timedelta(milliseconds=decision_lag)
                fill_time = np.datetime64(fill_time)
                while fill_time in asset_record_ar["index"]:
                    fill_time += np.timedelta64(1, "ms")

                wallet_balance = chunk_virtual_state.available_balance
                for symbol_key, location in chunk_virtual_state.positions.items():
                    if location.amount == 0:
                        continue
                    symbol_price = candle_data_ar[cycle][f"{symbol_key}/CLOSE"]
                    if math.isnan(symbol_price):
                        continue
                    current_margin = abs(location.amount) * location.entry_price
                    wallet_balance += current_margin

                margin_ratio = abs(amount_shift) * open_price / wallet_balance

                order_id = random.randint(ORDER_ID_MIN, ORDER_ID_MAX)

                if amount_shift == 0.0:
                    raise ValueError("Amount of asset shift cannot be zero")

                if role is None:
                    raise ValueError("No trade role was specified")

                if fill_price <= 0:
                    raise ValueError("The fill price should be bigger than zero")

                original_size = asset_record_ar.shape[0]
                asset_record_ar.resize(original_size + 1)
                asset_record_ar[-1]["index"] = fill_time
                asset_record_ar[-1]["CAUSE"] = "AUTO_TRADE"
                asset_record_ar[-1]["SYMBOL"] = symbol
                asset_record_ar[-1]["SIDE"] = "BUY" if amount_shift > 0.0 else "SELL"
                asset_record_ar[-1]["FILL_PRICE"] = fill_price
                asset_record_ar[-1]["ROLE"] = role.value
                asset_record_ar[-1]["MARGIN_RATIO"] = margin_ratio
                asset_record_ar[-1]["ORDER_ID"] = order_id
                asset_record_ar[-1]["RESULT_ASSET"] = wallet_balance

                update_time = fill_time.astype(datetime).replace(tzinfo=timezone.utc)
                chunk_account_state.positions[symbol].update_time = update_time

        # ■■■■■ understand the situation ■■■■■

        wallet_balance = chunk_virtual_state.available_balance
        unrealized_profit = 0
        for symbol_key, location in chunk_virtual_state.positions.items():
            if location.amount == 0:
                continue
            symbol_price = candle_data_ar[cycle][f"{symbol_key}/CLOSE"]
            if math.isnan(symbol_price):
                continue
            current_margin = abs(location.amount) * location.entry_price
            wallet_balance += current_margin
            # assume that mark price doesn't wobble more than 5%
            key_open_price = candle_data_ar[cycle][f"{symbol_key}/OPEN"]
            key_close_price = candle_data_ar[cycle][f"{symbol_key}/CLOSE"]
            if location.amount < 0:
                basic_price = max(key_open_price, key_close_price) * 1.05
                key_high_price = candle_data_ar[cycle][f"{symbol_key}/HIGH"]
                extreme_price = min(basic_price, key_high_price)
            else:
                basic_price = min(key_open_price, key_close_price) * 0.95
                key_low_price = candle_data_ar[cycle][f"{symbol_key}/LOW"]
                extreme_price = max(basic_price, key_low_price)
            price_difference = extreme_price - location.entry_price
            unrealized_profit += price_difference * location.amount
        unrealized_change = unrealized_profit / wallet_balance

        # ■■■■■ update the account state (symbol independent) ■■■■■

        chunk_account_state.observed_until = current_moment
        chunk_account_state.wallet_balance = wallet_balance

        # ■■■■■ record (symbol independent) ■■■■■

        original_size = chunk_unrealized_changes_ar.shape[0]
        chunk_unrealized_changes_ar.resize(original_size + 1)
        chunk_unrealized_changes_ar[-1]["index"] = before_moment
        chunk_unrealized_changes_ar[-1]["0"] = unrealized_change

        # ■■■■■ make decision and place order ■■■■■

        record_row: np.record = candle_data_ar[cycle]
        current_candle_data = {
            k: float(record_row[k])
            for k in record_row.dtype.names or ()
            if k != "index"
        }
        record_row: np.record = indicators_ar[cycle]
        current_indicators = {
            k: float(record_row[k])
            for k in record_row.dtype.names or ()
            if k != "index"
        }

        decisions = make_decisions(
            strategy=strategy,
            target_symbols=target_symbols,
            current_moment=current_moment,
            current_candle_data=current_candle_data,
            current_indicators=current_indicators,
            account_state=chunk_account_state.model_copy(deep=True),
            scribbles=chunk_scribbles,
        )

        for symbol_key, symbol_decisions in decisions.items():
            for order_type, decision in symbol_decisions.items():
                placement = VirtualPlacement(
                    order_id=random.randint(ORDER_ID_MIN, ORDER_ID_MAX),
                    boundary=decision.boundary,
                    margin=decision.margin,
                )
                chunk_virtual_state.placements[symbol_key][order_type] = placement

        # ■■■■■ report the progress in seconds ■■■■■

        progress_in_time: pd.Timedelta = current_moment - first_calculation_moment
        progress_in_seconds = progress_in_time.total_seconds()
        if progress_in_seconds % 3600 == 0:
            # Do NOT report the progress too often for the sake of performance
            progress_list[target_progress] = max(progress_in_seconds, 0)

    # ■■■■■ convert back numpy objects to pandas objects ■■■■■

    chunk_asset_record = pd.DataFrame(asset_record_ar)
    chunk_asset_record = chunk_asset_record.set_index("index")
    chunk_asset_record.index.name = None
    chunk_asset_record.index = pd.to_datetime(chunk_asset_record.index, utc=True)

    chunk_unrealized_changes_df = pd.DataFrame(chunk_unrealized_changes_ar)
    chunk_unrealized_changes_df = chunk_unrealized_changes_df.set_index("index")
    chunk_unrealized_changes_df.index.name = None
    chunk_unrealized_changes_df.index = pd.to_datetime(
        chunk_unrealized_changes_df.index, utc=True
    )
    chunk_unrealized_changes = chunk_unrealized_changes_df["0"]

    # ■■■■■ return calculated data ■■■■■

    calculation_output = CalculationOutput(
        chunk_asset_record=chunk_asset_record,
        chunk_unrealized_changes=chunk_unrealized_changes,
        chunk_scribbles=chunk_scribbles,
        chunk_account_state=chunk_account_state,
        chunk_virtual_state=chunk_virtual_state,
    )
    return calculation_output



from PySide6.QtGui import QFont
from PySide6.QtWidgets import QLabel, QWidget


class BrandLabel(QLabel):
    def __init__(self, parent: QWidget, text="", size=9):
        super().__init__(parent)
        font = QFont("Lexend", size)
        font.setWeight(QFont.Weight.Bold)
        self.setFont(font)
        self.setText(text)



import logging
import math
import pickle
import webbrowser
from asyncio import gather, sleep, wait
from collections.abc import Coroutine
from datetime import datetime, timedelta, timezone
from typing import Any

import aiofiles
import aiofiles.os
import numpy as np
import pandas as pd
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from PySide6.QtWidgets import QMenu

from solie.common import UniqueTask, outsource, spawn, spawn_blocking
from solie.overlay import LongTextView
from solie.utility import (
    ApiRequester,
    ApiRequestError,
    ApiStreamer,
    BookTicker,
    Decision,
    DurationRecorder,
    MarkPrice,
    OpenOrder,
    OrderType,
    PositionDirection,
    RWLock,
    TransactionSettings,
    ball_ceil,
    create_empty_account_state,
    create_empty_asset_record,
    create_empty_unrealized_changes,
    internet_connected,
    list_to_dict,
    make_decisions,
    make_indicators,
    slice_deque,
    sort_data_frame,
    sort_series,
    to_moment,
    when_internet_connected,
    when_internet_disconnected,
)
from solie.widget import ask, overlay
from solie.window import Window

from .united import team

logger = logging.getLogger(__name__)


class Transactor:
    def __init__(self, window: Window, scheduler: AsyncIOScheduler):
        # ■■■■■ for data management ■■■■■

        self.window = window
        self.scheduler = scheduler
        self.workerpath = window.datapath / "transactor"

        # ■■■■■ internal memory ■■■■■

        self.line_display_task = UniqueTask()
        self.range_display_task = UniqueTask()

        self.maximum_quantities: dict[str, float] = {}  # Symbol and value
        self.minimum_notionals: dict[str, float] = {}  # Symbol and value
        self.price_precisions: dict[str, int] = {}  # Symbol and decimal places
        self.quantity_precisions: dict[str, int] = {}  # Symbol and decimal places
        self.maximum_leverages: dict[str, int] = {}  # Symbol and value
        self.leverages: dict[str, int] = {}  # Symbol and value
        self.is_key_restrictions_satisfied = True

        # ■■■■■ remember and display ■■■■■

        self.api_requester = ApiRequester()

        self.viewing_symbol = window.data_settings.target_symbols[0]
        self.should_draw_frequently = True

        self.account_state = create_empty_account_state(
            window.data_settings.target_symbols
        )

        self.scribbles: dict[Any, Any] = {}
        self.transaction_settings = TransactionSettings()
        self.unrealized_changes = RWLock(create_empty_unrealized_changes())
        self.asset_record = RWLock(create_empty_asset_record())
        self.auto_order_record = RWLock(
            pd.DataFrame(
                columns=["SYMBOL", "ORDER_ID"],
                index=pd.DatetimeIndex([], tz="UTC"),
            )
        )

        # ■■■■■ repetitive schedules ■■■■■

        self.scheduler.add_job(
            self.display_status_information,
            trigger="cron",
            second="*",
        )
        self.scheduler.add_job(
            self.display_range_information,
            trigger="cron",
            second="*",
        )
        self.scheduler.add_job(
            self.cancel_conflicting_orders,
            trigger="cron",
            second="*",
        )
        self.scheduler.add_job(
            self.display_lines,
            trigger="cron",
            second="*/10",
            kwargs={"periodic": True, "frequent": True},
        )
        self.scheduler.add_job(
            self.pan_view_range,
            trigger="cron",
            second="*/10",
        )
        self.scheduler.add_job(
            self.perform_transaction,
            trigger="cron",
            second="*/10",
        )
        self.scheduler.add_job(
            self.save_scribbles,
            trigger="cron",
            second="*/10",
        )
        self.scheduler.add_job(
            self.watch_binance,
            trigger="cron",
            second="*/10",
        )
        self.scheduler.add_job(
            self.organize_data,
            trigger="cron",
            minute="*",
        )
        self.scheduler.add_job(
            self.save_large_data,
            trigger="cron",
            hour="*",
        )
        self.scheduler.add_job(
            self.update_user_data_stream,
            trigger="cron",
            hour="*",
        )

        # ■■■■■ websocket streamings ■■■■■

        self.user_data_streamer: ApiStreamer | None = None

        # ■■■■■ invoked by the internet connection status change ■■■■■

        when_internet_connected(self.watch_binance)
        when_internet_connected(self.update_user_data_stream)
        when_internet_disconnected(self.update_user_data_stream)

        # ■■■■■ connect UI events ■■■■■

        # Special widgets
        job = self.display_range_information
        outsource(window.transaction_graph.price_widget.sigRangeChanged, job)
        job = self.set_minimum_view_range
        outsource(window.transaction_graph.price_widget.sigRangeChanged, job)
        job = self.update_automation_settings
        outsource(window.comboBox_2.currentIndexChanged, job)
        job = self.update_automation_settings
        outsource(window.checkBox.toggled, job)
        job = self.update_keys
        outsource(window.lineEdit_4.editingFinished, job)
        job = self.update_keys
        outsource(window.lineEdit_6.editingFinished, job)
        job = self.toggle_frequent_draw
        outsource(window.checkBox_2.toggled, job)
        job = self.display_day_range
        outsource(window.pushButton_14.clicked, job)
        job = self.update_mode_settings
        outsource(window.spinBox.editingFinished, job)
        job = self.update_viewing_symbol
        outsource(window.comboBox_4.currentIndexChanged, job)

        action_menu = QMenu(self.window)
        self.window.pushButton_12.setMenu(action_menu)

        text = "Open Binance exchange"
        job = self.open_exchange
        new_action = action_menu.addAction(text)
        outsource(new_action.triggered, job)
        text = "Open Binance futures wallet"
        job = self.open_futures_wallet_page
        new_action = action_menu.addAction(text)
        outsource(new_action.triggered, job)
        text = "Open Binance API management webpage"
        job = self.open_api_management_page
        new_action = action_menu.addAction(text)
        outsource(new_action.triggered, job)
        text = "Clear all positions and open orders"
        job = self.clear_positions_and_open_orders
        new_action = action_menu.addAction(text)
        outsource(new_action.triggered, job)
        text = "Display same range as simulation graph"
        job = self.match_graph_range
        new_action = action_menu.addAction(text)
        outsource(new_action.triggered, job)
        text = "Show Raw Account State Object"
        job = self.show_raw_account_state_object
        new_action = action_menu.addAction(text)
        outsource(new_action.triggered, job)

    async def load(self):
        await aiofiles.os.makedirs(self.workerpath, exist_ok=True)

        # scribbles
        filepath = self.workerpath / "scribbles.pickle"
        if await aiofiles.os.path.isfile(filepath):
            async with aiofiles.open(filepath, "rb") as file:
                content = await file.read()
                self.scribbles = pickle.loads(content)

        # transaction settings
        filepath = self.workerpath / "transaction_settings.json"
        if await aiofiles.os.path.isfile(filepath):
            async with aiofiles.open(filepath, "r", encoding="utf8") as file:
                read_data = TransactionSettings.model_validate_json(await file.read())
            self.transaction_settings = read_data
            state = read_data.should_transact
            self.window.checkBox.setChecked(state)
            strategy_index = read_data.strategy_index
            self.window.comboBox_2.setCurrentIndex(strategy_index)
            new_value = read_data.desired_leverage
            self.window.spinBox.setValue(new_value)
            text = read_data.binance_api_key
            self.window.lineEdit_4.setText(text)
            text = read_data.binance_api_secret
            self.window.lineEdit_6.setText(text)
            self.api_requester.update_keys(
                read_data.binance_api_key, read_data.binance_api_secret
            )

        # unrealized changes
        filepath = self.workerpath / "unrealized_changes.pickle"
        if await aiofiles.os.path.isfile(filepath):
            sr: pd.Series = await spawn_blocking(pd.read_pickle, filepath)
            self.unrealized_changes = RWLock(sr)

        # asset record
        filepath = self.workerpath / "asset_record.pickle"
        if await aiofiles.os.path.isfile(filepath):
            df: pd.DataFrame = await spawn_blocking(pd.read_pickle, filepath)
            self.asset_record = RWLock(df)

        # auto order record
        filepath = self.workerpath / "auto_order_record.pickle"
        if await aiofiles.os.path.isfile(filepath):
            df: pd.DataFrame = await spawn_blocking(pd.read_pickle, filepath)
            self.auto_order_record = RWLock(df)

    async def organize_data(self):
        async with self.unrealized_changes.write_lock as cell:
            if not cell.data.index.is_unique:
                unique_index = cell.data.index.drop_duplicates()
                cell.data = cell.data.reindex(unique_index)
            if not cell.data.index.is_monotonic_increasing:
                cell.data = await spawn_blocking(sort_series, cell.data)

        async with self.auto_order_record.write_lock as cell:
            if not cell.data.index.is_unique:
                unique_index = cell.data.index.drop_duplicates()
                cell.data = cell.data.reindex(unique_index)
            if not cell.data.index.is_monotonic_increasing:
                cell.data = await spawn_blocking(sort_data_frame, cell.data)
            max_length = 2**16
            if len(cell.data) > max_length:
                cell.data = cell.data.iloc[-max_length:].copy()

        async with self.asset_record.write_lock as cell:
            if not cell.data.index.is_unique:
                unique_index = cell.data.index.drop_duplicates()
                cell.data = cell.data.reindex(unique_index)
            if not cell.data.index.is_monotonic_increasing:
                cell.data = await spawn_blocking(sort_data_frame, cell.data)

    async def save_large_data(self):
        async with self.unrealized_changes.read_lock as cell:
            unrealized_changes = cell.data.copy()
        await spawn_blocking(
            unrealized_changes.to_pickle,
            self.workerpath / "unrealized_changes.pickle",
        )

        async with self.auto_order_record.read_lock as cell:
            auto_order_record = cell.data.copy()
        await spawn_blocking(
            auto_order_record.to_pickle,
            self.workerpath / "auto_order_record.pickle",
        )

        async with self.asset_record.read_lock as cell:
            asset_record = cell.data.copy()
        await spawn_blocking(
            asset_record.to_pickle,
            self.workerpath / "asset_record.pickle",
        )

    async def save_scribbles(self):
        filepath = self.workerpath / "scribbles.pickle"
        async with aiofiles.open(filepath, "wb") as file:
            content = pickle.dumps(self.scribbles)
            await file.write(content)

    async def update_user_data_stream(self):
        """
        Prepares the WebSocket user data stream from Binance,
        providing updates on account changes and market order results.

        Although rare, the listen key may change over time.
        Additionally, the Binance API documentation recommends
        extending the user data stream every hour.
        Thus, this function should be called periodically to maintain the stream.

        - https://binance-docs.github.io/apidocs/futures/en/#start-user-data-stream-user_stream
        """

        async def close_stream():
            if self.user_data_streamer:
                await self.user_data_streamer.close()
                self.user_data_streamer = None

        if not internet_connected():
            await close_stream()
            return

        try:
            response = await self.api_requester.binance(
                http_method="POST",
                path="/fapi/v1/listenKey",
            )
        except ApiRequestError:
            await close_stream()
            return

        listen_key = response["listenKey"]
        new_url = f"wss://fstream.binance.com/ws/{listen_key}"

        if self.user_data_streamer:
            if new_url == self.user_data_streamer.url:
                # If the listen key hasn't changed, do nothing.
                return
            else:
                # If the listen key has changed, close the previous session.
                await self.user_data_streamer.close()

        self.user_data_streamer = ApiStreamer(
            new_url,
            self.listen_to_account,
        )

    async def listen_to_account(self, received: dict[str, Any]):
        # ■■■■■ default values ■■■■■

        event_type = str(received["e"])
        event_timestamp = int(received["E"]) / 1000
        event_time = datetime.fromtimestamp(event_timestamp, tz=timezone.utc)

        self.account_state.observed_until = event_time

        # ■■■■■ do the task according to event type ■■■■■

        if event_type == "listenKeyExpired":
            text = "Binance user data stream listen key has expired"
            logger.warning(text)
            await self.update_user_data_stream()

        elif event_type == "ACCOUNT_UPDATE":
            about_update = received["a"]
            about_assets = about_update["B"]
            about_positions = about_update["P"]

            asset_token = self.window.data_settings.asset_token

            about_assets_keyed = list_to_dict(about_assets, "a")
            about_asset = about_assets_keyed[asset_token]
            wallet_balance = float(about_asset["wb"])
            self.account_state.wallet_balance = wallet_balance

            about_positions_keyed = list_to_dict(about_positions, "ps")
            if "BOTH" in about_positions_keyed:
                about_position = about_positions_keyed["BOTH"]

                target_symbols = self.window.data_settings.target_symbols
                if about_position["s"] not in target_symbols:
                    return

                symbol = str(about_position["s"])
                amount = float(about_position["pa"])
                entry_price = float(about_position["ep"])

                leverage = self.leverages[symbol]
                margin = abs(amount) * entry_price / leverage
                if amount < 0.0:
                    direction = PositionDirection.SHORT
                elif amount > 0.0:
                    direction = PositionDirection.LONG
                else:
                    direction = PositionDirection.LONG

                position = self.account_state.positions[symbol]
                position.margin = margin
                position.direction = direction
                position.entry_price = entry_price
                position.update_time = event_time

        elif event_type == "ORDER_TRADE_UPDATE":
            about_update = received["o"]

            target_symbols = self.window.data_settings.target_symbols
            if about_update["s"] not in target_symbols:
                return

            # from received
            symbol = str(about_update["s"])
            order_id = int(about_update["i"])
            order_type = str(about_update["o"])
            order_status = str(about_update["X"])
            execution_type = str(about_update["x"])

            side = str(about_update["S"])
            close_position = bool(about_update["cp"])
            is_maker = bool(about_update["m"])

            origianal_quantity = float(about_update["q"])
            executed_quantity = float(about_update["z"])
            last_filled_quantity = float(about_update["l"])
            last_filled_price = float(about_update["L"])
            price = float(about_update["p"])
            stop_price = float(about_update["sp"])
            commission = float(about_update["n"])
            realized_profit = float(about_update["rp"])

            # from remembered
            leverage = self.leverages[symbol]
            wallet_balance = self.account_state.wallet_balance

            # when the order is removed
            if order_status not in ("NEW", "PARTIALLY_FILLED"):
                if order_id in self.account_state.open_orders[symbol].keys():
                    self.account_state.open_orders[symbol].pop(order_id)

            # when the order is left or created
            if order_status in ("NEW", "PARTIALLY_FILLED"):
                if order_type == "STOP_MARKET":
                    if close_position:
                        if side == "BUY":
                            order_type = OrderType.LATER_UP_CLOSE
                            boundary = stop_price
                            left_margin = None
                        elif side == "SELL":
                            order_type = OrderType.LATER_DOWN_CLOSE
                            boundary = stop_price
                            left_margin = None
                        else:
                            raise ValueError("Cannot order with this side")
                    elif side == "BUY":
                        order_type = OrderType.LATER_UP_BUY
                        boundary = stop_price
                        left_quantity = origianal_quantity
                        left_margin = left_quantity * boundary / leverage
                    elif side == "SELL":
                        order_type = OrderType.LATER_DOWN_SELL
                        boundary = stop_price
                        left_quantity = origianal_quantity
                        left_margin = left_quantity * boundary / leverage
                    else:
                        raise ValueError("Cannot order with this side")
                elif order_type == "TAKE_PROFIT_MARKET":
                    if close_position:
                        if side == "BUY":
                            order_type = OrderType.LATER_DOWN_CLOSE
                            boundary = stop_price
                            left_margin = None
                        elif side == "SELL":
                            order_type = OrderType.LATER_UP_CLOSE
                            boundary = stop_price
                            left_margin = None
                        else:
                            raise ValueError("Cannot order with this side")
                    elif side == "BUY":
                        order_type = OrderType.LATER_DOWN_BUY
                        boundary = stop_price
                        left_quantity = origianal_quantity
                        left_margin = left_quantity * boundary / leverage
                    elif side == "SELL":
                        order_type = OrderType.LATER_UP_SELL
                        boundary = stop_price
                        left_quantity = origianal_quantity
                        left_margin = left_quantity * boundary / leverage
                    else:
                        raise ValueError("Cannot order with this side")
                elif order_type == "LIMIT":
                    if side == "BUY":
                        order_type = OrderType.BOOK_BUY
                        boundary = price
                        left_quantity = origianal_quantity - executed_quantity
                        left_margin = left_quantity * boundary / leverage
                    elif side == "SELL":
                        order_type = OrderType.BOOK_SELL
                        boundary = price
                        left_quantity = origianal_quantity - executed_quantity
                        left_margin = left_quantity * boundary / leverage
                    else:
                        raise ValueError("Cannot order with this side")
                else:
                    order_type = OrderType.OTHER
                    boundary = max(price, stop_price)
                    left_quantity = origianal_quantity - executed_quantity
                    left_margin = left_quantity * boundary / leverage

                self.account_state.open_orders[symbol][order_id] = OpenOrder(
                    order_type=order_type,
                    boundary=boundary,
                    left_margin=left_margin,
                )

            # when the order is filled
            if execution_type == "TRADE":
                added_revenue = realized_profit - commission
                added_notional = last_filled_price * last_filled_quantity
                added_margin = added_notional / leverage
                added_margin_ratio = added_margin / wallet_balance

                async with self.auto_order_record.read_lock as cell:
                    symbol_df = cell.data[cell.data["SYMBOL"] == symbol]
                    unique_order_ids = symbol_df["ORDER_ID"].unique()

                async with self.asset_record.write_lock as cell:
                    symbol_df = cell.data[cell.data["SYMBOL"] == symbol]
                    recorded_id_list = symbol_df["ORDER_ID"].tolist()
                    does_record_exist = order_id in recorded_id_list
                    last_index = cell.data.index[-1]
                    if does_record_exist:
                        mask_sr = symbol_df["ORDER_ID"] == order_id
                        rec_time = symbol_df.index[mask_sr][0]
                        rec_value = float(symbol_df.loc[rec_time, "MARGIN_RATIO"])  # type:ignore
                        new_value = rec_value + added_margin_ratio
                        cell.data.loc[rec_time, "MARGIN_RATIO"] = new_value
                        last_asset = float(cell.data.loc[last_index, "RESULT_ASSET"])  # type:ignore
                        new_value = last_asset + added_revenue
                        cell.data.loc[last_index, "RESULT_ASSET"] = new_value
                    else:
                        record_time = event_time
                        while record_time in cell.data.index:
                            record_time += timedelta(milliseconds=1)
                        new_value = symbol
                        cell.data.loc[record_time, "SYMBOL"] = new_value
                        new_value = "SELL" if side == "SELL" else "BUY"
                        cell.data.loc[record_time, "SIDE"] = new_value
                        new_value = last_filled_price
                        cell.data.loc[record_time, "FILL_PRICE"] = new_value
                        new_value = "MAKER" if is_maker else "TAKER"
                        cell.data.loc[record_time, "ROLE"] = new_value
                        new_value = added_margin_ratio
                        cell.data.loc[record_time, "MARGIN_RATIO"] = new_value
                        new_value = order_id
                        cell.data.loc[record_time, "ORDER_ID"] = new_value
                        last_asset = float(cell.data.loc[last_index, "RESULT_ASSET"])  # type:ignore
                        new_value = last_asset + added_revenue
                        cell.data.loc[record_time, "RESULT_ASSET"] = new_value
                        if order_id in unique_order_ids:
                            cell.data.loc[record_time, "CAUSE"] = "AUTO_TRADE"
                        else:
                            cell.data.loc[record_time, "CAUSE"] = "MANUAL_TRADE"
                    if not cell.data.index.is_monotonic_increasing:
                        cell.data = await spawn_blocking(sort_data_frame, cell.data)

        # ■■■■■ cancel conflicting orders ■■■■■

        await self.cancel_conflicting_orders()

    async def open_exchange(self):
        symbol = self.viewing_symbol
        await spawn_blocking(
            webbrowser.open,
            f"https://www.binance.com/en/futures/{symbol}",
        )

    async def open_futures_wallet_page(self):
        await spawn_blocking(
            webbrowser.open,
            "https://www.binance.com/en/my/wallet/account/futures",
        )

    async def open_api_management_page(self):
        await spawn_blocking(
            webbrowser.open,
            "https://www.binance.com/en/my/settings/api-management",
        )

    async def save_transaction_settings(self):
        filepath = self.workerpath / "transaction_settings.json"
        async with aiofiles.open(filepath, "w", encoding="utf8") as file:
            await file.write(self.transaction_settings.model_dump_json(indent=2))

    async def update_keys(self):
        binance_api_key = self.window.lineEdit_4.text()
        binance_api_secret = self.window.lineEdit_6.text()

        self.transaction_settings.binance_api_key = binance_api_key
        self.transaction_settings.binance_api_secret = binance_api_secret

        await self.save_transaction_settings()
        self.api_requester.update_keys(binance_api_key, binance_api_secret)
        await self.update_user_data_stream()

    async def update_automation_settings(self):
        # ■■■■■ get information about strategy ■■■■■

        strategy_index = self.window.comboBox_2.currentIndex()
        self.transaction_settings.strategy_index = strategy_index

        spawn(self.display_lines())

        # ■■■■■ is automation turned on ■■■■■

        is_checked = self.window.checkBox.isChecked()

        if is_checked:
            self.transaction_settings.should_transact = True
        else:
            self.transaction_settings.should_transact = False

        # ■■■■■ save ■■■■■

        await self.save_transaction_settings()

    async def display_range_information(self):
        self.range_display_task.spawn(self._display_range_information())

    async def _display_range_information(self):
        symbol = self.viewing_symbol
        price_widget = self.window.transaction_graph.price_widget

        range_start_timestamp = price_widget.getAxis("bottom").range[0]
        range_start_timestamp = max(range_start_timestamp, 0.0)
        range_start = datetime.fromtimestamp(range_start_timestamp, tz=timezone.utc)

        range_end_timestamp = price_widget.getAxis("bottom").range[1]
        if range_end_timestamp < 0.0:
            # case when pyqtgraph passed negative value because it's too big
            range_end_timestamp = 9223339636.0
        else:
            # maximum value available in pandas
            range_end_timestamp = min(range_end_timestamp, 9223339636.0)
        range_end = datetime.fromtimestamp(range_end_timestamp, tz=timezone.utc)

        range_length = range_end - range_start
        range_days = range_length.days
        range_hours, remains = divmod(range_length.seconds, 3600)
        range_minutes, remains = divmod(remains, 60)

        async with self.unrealized_changes.read_lock as cell:
            unrealized_changes = cell.data[range_start:range_end].copy()
        async with self.asset_record.read_lock as cell:
            asset_record = cell.data[range_start:range_end].copy()

        auto_trade_mask = asset_record["CAUSE"] == "AUTO_TRADE"
        asset_changes = asset_record["RESULT_ASSET"].pct_change(fill_method=None) + 1  # type:ignore
        asset_record = asset_record[auto_trade_mask]
        asset_changes = asset_changes.reindex(asset_record.index, fill_value=1.0)
        symbol_mask = asset_record["SYMBOL"] == symbol

        # trade count
        total_change_count = len(asset_changes)
        symbol_change_count = len(asset_changes[symbol_mask])
        # trade volume
        if len(asset_record) > 0:
            total_margin_ratio = asset_record["MARGIN_RATIO"].sum()
        else:
            total_margin_ratio = 0
        if len(asset_record[symbol_mask]) > 0:
            symbol_margin_ratio = asset_record[symbol_mask]["MARGIN_RATIO"].sum()
        else:
            symbol_margin_ratio = 0
        # asset changes
        if len(asset_changes) > 0:
            total_yield = asset_changes.cumprod().iloc[-1]
            total_yield = (total_yield - 1) * 100
        else:
            total_yield = 0
        if len(asset_changes[symbol_mask]) > 0:
            symbol_yield = asset_changes[symbol_mask].cumprod().iloc[-1]
            symbol_yield = (symbol_yield - 1) * 100
        else:
            symbol_yield = 0
        # least unrealized changes
        if len(unrealized_changes) > 0:
            min_unrealized_change = unrealized_changes.min()
        else:
            min_unrealized_change = 0

        view_range = price_widget.getAxis("left").range
        range_down = view_range[0]
        range_up = view_range[1]
        price_range_height = (1 - range_down / range_up) * 100

        text = ""
        text += f"Visible time range {range_days}d {range_hours}h {range_minutes}s"
        text += "  ⦁  "
        text += "Visible price range"
        text += f" {price_range_height:.2f}%"
        text += "  ⦁  "
        text += f"Transaction count {symbol_change_count}({total_change_count})"
        text += "  ⦁  "
        text += "Transaction amount"
        text += f" *{symbol_margin_ratio:.4f}({total_margin_ratio:.4f})"
        text += "  ⦁  "
        text += f"Total realized profit {symbol_yield:+.4f}({total_yield:+.4f})%"
        text += "  ⦁  "
        text += "Lowest unrealized profit"
        text += f" {min_unrealized_change * 100:+.4f}%"
        self.window.label_8.setText(text)

    async def set_minimum_view_range(self):
        widget = self.window.transaction_graph.price_widget
        range_down = widget.getAxis("left").range[0]
        widget.plotItem.vb.setLimits(minYRange=range_down * 0.005)  # type:ignore
        widget = self.window.transaction_graph.asset_widget
        range_down = widget.getAxis("left").range[0]
        widget.plotItem.vb.setLimits(minYRange=range_down * 0.005)  # type:ignore

    async def display_strategy_index(self):
        strategy_index = self.transaction_settings.strategy_index
        self.window.comboBox_2.setCurrentIndex(strategy_index)

    async def display_lines(self, periodic=False, frequent=False):
        self.line_display_task.spawn(self._display_lines(periodic, frequent))

    async def _display_lines(self, periodic: bool, frequent: bool):
        # ■■■■■ get basic information ■■■■■

        symbol = self.viewing_symbol
        strategy_index = self.transaction_settings.strategy_index
        strategy = team.strategist.strategies[strategy_index]

        should_draw_frequently = self.should_draw_frequently
        if frequent and not should_draw_frequently:
            return

        # ■■■■■ ensure that the latest data was added ■■■■■

        async with team.collector.candle_data.read_lock as cell:
            if len(cell.data) == 0:
                return

        current_moment = to_moment(datetime.now(timezone.utc))
        before_moment = current_moment - timedelta(seconds=10)

        if periodic:
            for _ in range(50):
                async with team.collector.candle_data.read_lock as cell:
                    last_index = cell.data.index[-1]
                    if last_index == before_moment:
                        break
                await sleep(0.1)

        # ■■■■■ get ready for task duration measurement ■■■■■

        duration_recorder = DurationRecorder("DISPLAY_TRANSACTION_LINES")

        # ■■■■■ set the slicing range ■■■■■

        if should_draw_frequently:
            get_from = datetime.now(timezone.utc) - timedelta(days=28)
            slice_from = datetime.now(timezone.utc) - timedelta(hours=24)
            slice_until = datetime.now(timezone.utc)
        else:
            current_year = datetime.now(timezone.utc).year
            get_from = datetime(current_year, 1, 1, tzinfo=timezone.utc)
            slice_from = datetime(current_year, 1, 1, tzinfo=timezone.utc)
            slice_until = datetime.now(timezone.utc)
        slice_until -= timedelta(seconds=1)

        # ■■■■■ draw light lines ■■■■■

        realtime_data = slice_deque(team.collector.realtime_data, 2 ** (10 + 6))
        mark_prices: list[MarkPrice] = []
        book_tickers: list[BookTicker] = []
        for realtime_record in realtime_data:
            if isinstance(realtime_record, BookTicker):
                if realtime_record.symbol == symbol:
                    book_tickers.append(realtime_record)
            else:
                is_valid = realtime_record.mark_price > 0.0
                if is_valid and realtime_record.symbol == symbol:
                    mark_prices.append(realtime_record)

        aggregate_trades = slice_deque(team.collector.aggregate_trades, 2 ** (10 + 6))
        aggregate_trades = [a for a in aggregate_trades if a.symbol == symbol]

        position = self.account_state.positions[symbol]
        if position.direction == PositionDirection.NONE:
            entry_price = None
        else:
            entry_price = position.entry_price

        await self.window.transaction_graph.update_light_lines(
            mark_prices=mark_prices,
            aggregate_trades=aggregate_trades,
            book_tickers=book_tickers,
            entry_price=entry_price,
            observed_until=self.account_state.observed_until,
        )

        # ■■■■■ draw heavy lines ■■■■■

        async with team.collector.candle_data.read_lock as cell:
            columns = [str(s) for s in cell.data.columns]
            chosen_columns = [s for s in columns if s.startswith(symbol)]
            candle_data_original = cell.data[chosen_columns][
                get_from:slice_until
            ].copy()
        async with self.unrealized_changes.read_lock as cell:
            unrealized_changes = cell.data.copy()
        async with self.asset_record.read_lock as cell:
            if len(cell.data) > 0:
                last_asset = cell.data.iloc[-1]["RESULT_ASSET"]
            else:
                last_asset = None
            before_record = cell.data[:slice_from]
            if len(before_record) > 0:
                before_asset = before_record.iloc[-1]["RESULT_ASSET"]
            else:
                before_asset = None
            asset_record = cell.data[slice_from:].copy()

        candle_data = candle_data_original[slice_from:]

        # add the right end
        if len(candle_data) > 0:
            last_written_moment = candle_data.index[-1]
            new_moment = last_written_moment + timedelta(seconds=10)
            new_index = candle_data.index.union([new_moment])
            candle_data = candle_data.reindex(new_index)

        if last_asset is not None:
            observed_until = self.account_state.observed_until
            if len(asset_record) == 0 or asset_record.index[-1] < observed_until:
                if slice_from < observed_until:
                    asset_record.loc[observed_until, "CAUSE"] = "OTHER"
                    asset_record.loc[observed_until, "RESULT_ASSET"] = last_asset
                    if not asset_record.index.is_monotonic_increasing:
                        asset_record = await spawn_blocking(
                            sort_data_frame, asset_record
                        )

        # add the left end
        if before_asset is not None:
            asset_record.loc[slice_from, "CAUSE"] = "OTHER"
            asset_record.loc[slice_from, "RESULT_ASSET"] = before_asset
            if not asset_record.index.is_monotonic_increasing:
                asset_record = await spawn_blocking(sort_data_frame, asset_record)

        await self.window.transaction_graph.update_heavy_lines(
            symbol=symbol,
            candle_data=candle_data,
            asset_record=asset_record,
            unrealized_changes=unrealized_changes,
        )

        # ■■■■■ draw custom lines ■■■■■

        indicators = await spawn_blocking(
            make_indicators,
            strategy=strategy,
            target_symbols=[self.viewing_symbol],
            candle_data=candle_data_original,
        )

        indicators = indicators[slice_from:slice_until]

        await self.window.transaction_graph.update_custom_lines(symbol, indicators)

        # ■■■■■ record task duration ■■■■■

        duration_recorder.record()

        # ■■■■■ set minimum view range ■■■■■

        await self.set_minimum_view_range()

    async def toggle_frequent_draw(self):
        is_checked = self.window.checkBox_2.isChecked()
        if is_checked:
            self.should_draw_frequently = True
        else:
            self.should_draw_frequently = False
        await self.display_lines()

    async def update_viewing_symbol(self):
        alias = self.window.comboBox_4.currentText()
        symbol = self.window.alias_to_symbol[alias]
        self.viewing_symbol = symbol

        spawn(self.display_lines())
        spawn(self.display_status_information())
        spawn(self.display_range_information())

    async def display_status_information(self):
        # ■■■■■ Display important things first ■■■■■

        time_passed = datetime.now(timezone.utc) - self.account_state.observed_until
        if time_passed > timedelta(seconds=30):
            text = (
                "Couldn't get the latest info on your Binance account due to a problem"
                " with your key or connection to the Binance server."
            )
            self.window.label_16.setText(text)
            return

        if not self.is_key_restrictions_satisfied:
            text = (
                "API key's restrictions are not satisfied. Auto transaction is"
                " disabled. Go to your Binance API managements webpage to change"
                " the restrictions."
            )
            self.window.label_16.setText(text)
            return

        cumulation_rate = await team.collector.check_candle_data_cumulation_rate()
        if cumulation_rate < 1:
            text = (
                "For auto transaction to work, the past 24 hour accumulation rate of"
                " candle data must be 100%. Auto transaction is disabled."
            )
            self.window.label_16.setText(text)
            return

        # ■■■■■ display assets and positions information ■■■■■

        position = self.account_state.positions[self.viewing_symbol]
        if position.direction == PositionDirection.LONG:
            direction_text = "long"
        elif position.direction == PositionDirection.SHORT:
            direction_text = "short"
        else:
            direction_text = "none"
        margin_sum = 0
        for each_position in self.account_state.positions.values():
            margin_sum += each_position.margin

        open_orders = self.account_state.open_orders
        open_orders_count = len(open_orders[self.viewing_symbol])
        all_open_orders_count = 0
        for symbol_open_orders in open_orders.values():
            all_open_orders_count += len(symbol_open_orders)

        text = ""
        text += "Total asset"
        text += f" ${self.account_state.wallet_balance:.4f}"
        text += "  ⦁  "
        text += f"Investment ${position.margin:.4f}"
        text += f"({margin_sum:.4f})"
        text += "  ⦁  "
        text += f"Direction {direction_text}"
        text += "  ⦁  "
        text += "Entry price"
        text += f" ${position.entry_price:.4f}"
        text += "  ⦁  "
        text += "Open orders"
        text += f" {open_orders_count}"
        text += f"({all_open_orders_count})"

        self.window.label_16.setText(text)

    async def perform_transaction(self):
        # ■■■■■ Clear the progress bar ■■■■

        self.window.progressBar_2.setValue(0)

        # ■■■■■ Stop if conditions are not met ■■■■

        if not internet_connected():
            return

        if not self.transaction_settings.should_transact:
            return

        if not self.is_key_restrictions_satisfied:
            return

        cumulation_rate = await team.collector.check_candle_data_cumulation_rate()
        if cumulation_rate < 1:
            return

        # ■■■■■ Moment ■■■■■

        duration_recorder = DurationRecorder("PERFORM_TRANSACTION")
        current_moment = to_moment(datetime.now(timezone.utc))
        before_moment = current_moment - timedelta(seconds=10)

        # ■■■■■ Play the progress bar ■■■■■

        is_cycle_done = False

        async def play_progress_bar():
            passed_time = timedelta(seconds=0)
            while passed_time < timedelta(seconds=10):
                passed_time = datetime.now(timezone.utc) - current_moment
                if not is_cycle_done:
                    new_value = int(passed_time / timedelta(seconds=10) * 1000)
                else:
                    before_value = self.window.progressBar_2.value()
                    remaining = 1000 - before_value
                    new_value = before_value + math.ceil(remaining * 0.2)
                self.window.progressBar_2.setValue(new_value)
                await sleep(0.01)

        spawn(play_progress_bar())

        # ■■■■■ Check if the data exists ■■■■■

        async with team.collector.candle_data.read_lock as cell:
            if len(cell.data) == 0:
                # case when the app is executed for the first time
                return

        # ■■■■■ Wait for the latest data to be added ■■■■■

        for _ in range(50):
            async with team.collector.candle_data.read_lock as cell:
                last_index = cell.data.index[-1]
                if last_index == before_moment:
                    break
            await sleep(0.1)

        # ■■■■■ Get the candle data ■■■■■

        slice_from = datetime.now(timezone.utc) - timedelta(days=28)
        async with team.collector.candle_data.read_lock as cell:
            candle_data = cell.data[slice_from:].copy()

        # ■■■■■ Make decision ■■■■■

        target_symbols = self.window.data_settings.target_symbols

        strategy_index = self.transaction_settings.strategy_index
        strategy = team.strategist.strategies[strategy_index]

        # Split the candle data by symbol before calculation to reduct UI lags
        columns = [str(s) for s in cell.data.columns]
        coroutines: list[Coroutine[Any, Any, pd.DataFrame]] = []
        for symbol in target_symbols:
            chosen_columns = [s for s in columns if s.startswith(symbol)]
            coroutines.append(
                spawn_blocking(
                    make_indicators,
                    strategy=strategy,
                    target_symbols=[symbol],
                    candle_data=candle_data[chosen_columns],
                )
            )
            await sleep(0.0)
        symbol_indicators = await gather(*coroutines)
        indicators = pd.concat(symbol_indicators, axis="columns")

        record_row: np.record = candle_data.tail(1).to_records()[-1]
        current_candle_data = {
            k: float(record_row[k])
            for k in record_row.dtype.names or ()
            if k != "index"
        }
        record_row: np.record = indicators.to_records()[-1]
        current_indicators = {
            k: float(record_row[k])
            for k in record_row.dtype.names or ()
            if k != "index"
        }

        decisions = make_decisions(
            strategy=strategy,
            target_symbols=target_symbols,
            current_moment=current_moment,
            current_candle_data=current_candle_data,
            current_indicators=current_indicators,
            account_state=self.account_state,
            scribbles=self.scribbles,
        )

        # ■■■■■ Record task duration ■■■■■

        is_cycle_done = True
        duration_recorder.record()

        # ■■■■■ Place order ■■■■■

        await self.place_orders(decisions)

    async def display_day_range(self):
        range_start = (datetime.now(timezone.utc) - timedelta(hours=24)).timestamp()
        range_end = datetime.now(timezone.utc).timestamp()
        widget = self.window.transaction_graph.price_widget
        widget.setXRange(range_start, range_end)

    async def match_graph_range(self):
        graph_from = self.window.simulation_graph.price_widget
        graph_to = self.window.transaction_graph.price_widget
        graph_range = graph_from.getAxis("bottom").range
        range_start = graph_range[0]
        range_end = graph_range[1]
        graph_to.setXRange(range_start, range_end, padding=0)  # type:ignore

    async def update_mode_settings(self):
        desired_leverage = self.window.spinBox.value()
        self.transaction_settings.desired_leverage = desired_leverage

        # ■■■■■ tell if some symbol's leverage cannot be set as desired ■■■■■

        target_symbols = self.window.data_settings.target_symbols
        target_max_leverages: dict[str, int] = {}
        for symbol in target_symbols:
            max_leverage = self.maximum_leverages.get(symbol, 125)
            target_max_leverages[symbol] = max_leverage
        lowest_max_leverage = min(target_max_leverages.values())

        if lowest_max_leverage < desired_leverage:
            answer = await ask(
                "Leverage on some symbols cannot be set as desired",
                "Binance has its own leverage limit per market. For some symbols,"
                " leverage will be set as high as it can be, but not as same as the"
                " value entered. Generally, situation gets safer in terms of lowest"
                " unrealized changes and profit turns out to be a bit lower than"
                " simulation prediction with the same leverage.",
                ["Show details", "Okay"],
            )
            if answer == 1:
                texts: list[str] = []
                for symbol, max_leverage in target_max_leverages.items():
                    texts.append(f"{symbol} {max_leverage}")
                text = "\n".join(texts)
                await ask(
                    "These are highest available leverages",
                    text,
                    ["Okay"],
                )

        # ■■■■■ save ■■■■■

        await self.save_transaction_settings()

    async def watch_binance(self):
        # ■■■■■ Basic data ■■■■■

        target_symbols = self.window.data_settings.target_symbols
        asset_token = self.window.data_settings.asset_token

        # ■■■■■ Check internet connection ■■■■■

        if not internet_connected():
            return

        # ■■■■■ Moment ■■■■■

        current_moment = to_moment(datetime.now(timezone.utc))
        before_moment = current_moment - timedelta(seconds=10)

        # ■■■■■ Request exchange information ■■■■■

        payload: dict[str, Any] = {}
        response = await self.api_requester.binance(
            http_method="GET",
            path="/fapi/v1/exchangeInfo",
            payload=payload,
        )
        about_exchange = response

        for about_symbol in about_exchange["symbols"]:
            symbol = about_symbol["symbol"]

            about_filters = about_symbol["filters"]
            about_filters_keyed = list_to_dict(about_filters, "filterType")

            minimum_notional = float(about_filters_keyed["MIN_NOTIONAL"]["notional"])
            self.minimum_notionals[symbol] = minimum_notional

            maximum_quantity = min(
                float(about_filters_keyed["LOT_SIZE"]["maxQty"]),
                float(about_filters_keyed["MARKET_LOT_SIZE"]["maxQty"]),
            )
            self.maximum_quantities[symbol] = maximum_quantity

            ticksize = float(about_filters_keyed["PRICE_FILTER"]["tickSize"])
            price_precision = int(math.log10(1 / ticksize))
            self.price_precisions[symbol] = price_precision

            stepsize = float(about_filters_keyed["LOT_SIZE"]["stepSize"])
            quantity_precision = int(math.log10(1 / stepsize))
            self.quantity_precisions[symbol] = quantity_precision

        # ■■■■■ Request leverage bracket information ■■■■■

        try:
            payload = {
                "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
            }
            response = await self.api_requester.binance(
                http_method="GET",
                path="/fapi/v1/leverageBracket",
                payload=payload,
            )
            about_brackets = response

            for about_bracket in about_brackets:
                symbol = about_bracket["symbol"]
                max_leverage = about_bracket["brackets"][0]["initialLeverage"]
                self.maximum_leverages[symbol] = max_leverage
        except ApiRequestError:
            # when the key is not ready
            return

        # ■■■■■ Request account information ■■■■■

        try:
            payload = {
                "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
            }
            response = await self.api_requester.binance(
                http_method="GET",
                path="/fapi/v2/account",
                payload=payload,
            )
            about_account = response
        except ApiRequestError:
            # when the key is not ready
            return

        about_open_orders: dict[str, list[dict[str, Any]]] = {}

        async def job(symbol):
            payload = {
                "symbol": symbol,
                "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
            }
            response = await self.api_requester.binance(
                http_method="GET",
                path="/fapi/v1/openOrders",
                payload=payload,
            )
            about_open_orders[symbol] = response

        tasks = [spawn(job(s)) for s in target_symbols]
        await wait(tasks)

        # ■■■■■ Update account state ■■■■■

        # observed until
        self.account_state.observed_until = current_moment

        # wallet_balance
        about_assets = about_account["assets"]
        about_assets_keyed = list_to_dict(about_assets, "asset")
        about_asset = about_assets_keyed[asset_token]
        wallet_balance = float(about_asset["walletBalance"])
        self.account_state.wallet_balance = wallet_balance

        about_positions = about_account["positions"]
        about_positions_keyed = list_to_dict(about_positions, "symbol")

        # positions
        for symbol in target_symbols:
            about_position = about_positions_keyed[symbol]

            if float(about_position["notional"]) > 0.0:
                direction = PositionDirection.LONG
            elif float(about_position["notional"]) < 0.0:
                direction = PositionDirection.SHORT
            else:
                direction = PositionDirection.NONE

            entry_price = float(about_position["entryPrice"])
            update_time = int(float(about_position["updateTime"]) / 1000)
            update_time = datetime.fromtimestamp(update_time, tz=timezone.utc)
            leverage = int(about_position["leverage"])
            amount = float(about_position["positionAmt"])
            margin = abs(amount) * entry_price / leverage

            position = self.account_state.positions[symbol]
            position.margin = margin
            position.direction = direction
            position.entry_price = entry_price
            position.update_time = update_time

        # open orders
        open_orders: dict[str, dict[int, OpenOrder]] = {}
        for symbol in target_symbols:
            open_orders[symbol] = {}

        for symbol in target_symbols:
            about_position = about_positions_keyed[symbol]
            leverage = int(about_position["leverage"])

            for about_open_order in about_open_orders[symbol]:
                order_id = about_open_order["orderId"]
                order_type = about_open_order["type"]

                side = about_open_order["side"]
                close_position = about_open_order["closePosition"]

                price = float(about_open_order["price"])
                stop_price = float(about_open_order["stopPrice"])
                origianal_quantity = float(about_open_order["origQty"])
                executed_quantity = float(about_open_order["executedQty"])

                if order_type == "STOP_MARKET":
                    if close_position:
                        if side == "BUY":
                            order_type = OrderType.LATER_UP_CLOSE
                            boundary = stop_price
                            left_margin = None
                        elif side == "SELL":
                            order_type = OrderType.LATER_DOWN_CLOSE
                            boundary = stop_price
                            left_margin = None
                        else:
                            raise ValueError("Cannot order with this side")
                    elif side == "BUY":
                        order_type = OrderType.LATER_UP_BUY
                        boundary = stop_price
                        left_quantity = origianal_quantity
                        left_margin = left_quantity * boundary / leverage
                    elif side == "SELL":
                        order_type = OrderType.LATER_DOWN_SELL
                        boundary = stop_price
                        left_quantity = origianal_quantity
                        left_margin = left_quantity * boundary / leverage
                    else:
                        raise ValueError("Cannot order with this side")
                elif order_type == "TAKE_PROFIT_MARKET":
                    if close_position:
                        if side == "BUY":
                            order_type = OrderType.LATER_DOWN_CLOSE
                            boundary = stop_price
                            left_margin = None
                        elif side == "SELL":
                            order_type = OrderType.LATER_UP_CLOSE
                            boundary = stop_price
                            left_margin = None
                        else:
                            raise ValueError("Cannot order with this side")
                    elif side == "BUY":
                        order_type = OrderType.LATER_DOWN_BUY
                        boundary = stop_price
                        left_quantity = origianal_quantity
                        left_margin = left_quantity * boundary / leverage
                    elif side == "SELL":
                        order_type = OrderType.LATER_UP_SELL
                        boundary = stop_price
                        left_quantity = origianal_quantity
                        left_margin = left_quantity * boundary / leverage
                    else:
                        raise ValueError("Cannot order with this side")
                elif order_type == "LIMIT":
                    if side == "BUY":
                        order_type = OrderType.BOOK_BUY
                        boundary = price
                        left_quantity = origianal_quantity - executed_quantity
                        left_margin = left_quantity * boundary / leverage
                    elif side == "SELL":
                        order_type = OrderType.BOOK_SELL
                        boundary = price
                        left_quantity = origianal_quantity - executed_quantity
                        left_margin = left_quantity * boundary / leverage
                    else:
                        raise ValueError("Cannot order with this side")
                else:
                    order_type = OrderType.OTHER
                    boundary = max(price, stop_price)
                    left_quantity = origianal_quantity - executed_quantity
                    left_margin = left_quantity * boundary / leverage

                open_orders[symbol][order_id] = OpenOrder(
                    order_type=order_type,
                    boundary=boundary,
                    left_margin=left_margin,
                )

        self.account_state.open_orders = open_orders

        # ■■■■■ Update hidden state ■■■■■

        for symbol in target_symbols:
            about_position = about_positions_keyed[symbol]
            leverage = int(about_position["leverage"])
            self.leverages[symbol] = leverage

        # ■■■■■ Record unrealized change ■■■■■

        # unrealized profit is not included in walletBalance
        wallet_balance = float(about_asset["walletBalance"])
        if wallet_balance != 0:
            unrealized_profit = float(about_asset["unrealizedProfit"])
            unrealized_change = unrealized_profit / wallet_balance
        else:
            unrealized_change = 0

        async with self.unrealized_changes.write_lock as cell:
            cell.data[before_moment] = unrealized_change
            if not cell.data.index.is_monotonic_increasing:
                cell.data = await spawn_blocking(sort_series, cell.data)

        # ■■■■■ Make an asset trace if it's blank ■■■■■

        async with self.asset_record.write_lock as cell:
            if len(cell.data) == 0:
                wallet_balance = float(about_asset["walletBalance"])
                current_time = datetime.now(timezone.utc)
                cell.data.loc[current_time, "CAUSE"] = "OTHER"
                cell.data.loc[current_time, "RESULT_ASSET"] = wallet_balance

        # ■■■■■ When the wallet balance changed for no good reason ■■■■■

        wallet_balance = float(about_asset["walletBalance"])

        async with self.asset_record.read_lock as cell:
            last_index = cell.data.index[-1]
            last_asset = float(cell.data.loc[last_index, "RESULT_ASSET"])  # type:ignore

        if wallet_balance == 0:
            pass
        elif abs(wallet_balance - last_asset) / wallet_balance > 10**-9:
            # when the difference is bigger than a billionth
            # referal fee, funding fee, wallet transfer, etc..
            async with self.asset_record.write_lock as cell:
                current_time = datetime.now(timezone.utc)
                cell.data.loc[current_time, "CAUSE"] = "OTHER"
                cell.data.loc[current_time, "RESULT_ASSET"] = wallet_balance
                if not cell.data.index.is_monotonic_increasing:
                    cell.data = await spawn_blocking(sort_data_frame, cell.data)
        else:
            # when the difference is small enough to consider as an numeric error
            async with self.asset_record.write_lock as cell:
                last_index = cell.data.index[-1]
                cell.data.loc[last_index, "RESULT_ASSET"] = wallet_balance

        # ■■■■■ Correct mode of the account market if automation is turned on ■■■■■

        if self.transaction_settings.should_transact:

            async def job_1(symbol):
                about_position = about_positions_keyed[symbol]
                current_leverage = int(about_position["leverage"])

                desired_leverage = self.transaction_settings.desired_leverage
                maximum_leverages = self.maximum_leverages
                max_leverage = maximum_leverages.get(symbol, 125)
                goal_leverage = min(desired_leverage, max_leverage)

                if current_leverage != goal_leverage:
                    timestamp = int(datetime.now(timezone.utc).timestamp() * 1000)
                    payload = {
                        "symbol": symbol,
                        "timestamp": timestamp,
                        "leverage": goal_leverage,
                    }
                    await self.api_requester.binance(
                        http_method="POST",
                        path="/fapi/v1/leverage",
                        payload=payload,
                    )

            tasks = [spawn(job_1(s)) for s in target_symbols]
            await wait(tasks)

            async def job_2(symbol):
                about_position = about_positions_keyed[symbol]

                isolated = about_position["isolated"]
                notional = float(about_position["notional"])

                if isolated:
                    # close position if exists
                    if notional != 0:
                        decisions = {symbol: {OrderType.NOW_CLOSE: Decision()}}
                        await self.place_orders(decisions)

                    # change to crossed margin mode
                    timestamp = int(datetime.now(timezone.utc).timestamp() * 1000)
                    payload = {
                        "symbol": symbol,
                        "timestamp": timestamp,
                        "marginType": "CROSSED",
                    }
                    await self.api_requester.binance(
                        http_method="POST",
                        path="/fapi/v1/marginType",
                        payload=payload,
                    )

            tasks = [spawn(job_2(s)) for s in target_symbols]
            await wait(tasks)

            try:
                timestamp = int(datetime.now(timezone.utc).timestamp() * 1000)
                payload = {
                    "timestamp": timestamp,
                    "multiAssetsMargin": "false",
                }
                await self.api_requester.binance(
                    http_method="POST",
                    path="/fapi/v1/multiAssetsMargin",
                    payload=payload,
                )
            except ApiRequestError:
                pass

            try:
                timestamp = int(datetime.now(timezone.utc).timestamp() * 1000)
                payload = {
                    "timestamp": timestamp,
                    "dualSidePosition": "false",
                }
                await self.api_requester.binance(
                    http_method="POST",
                    path="/fapi/v1/positionSide/dual",
                    payload=payload,
                )
            except ApiRequestError:
                pass

        # ■■■■■ check API key restrictions ■■■■■

        payload = {
            "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
        }
        response = await self.api_requester.binance(
            http_method="GET",
            path="/sapi/v1/account/apiRestrictions",
            payload=payload,
            server="spot",
        )
        api_restrictions = response

        is_satisfied = True
        enable_required_restrictions = [
            "enableFutures",
        ]
        for restriction_name in enable_required_restrictions:
            is_enabled = api_restrictions[restriction_name]
            if not is_enabled:
                is_satisfied = False
        self.is_key_restrictions_satisfied = is_satisfied

    async def place_orders(self, decisions: dict[str, dict[OrderType, Decision]]):
        target_symbols = self.window.data_settings.target_symbols
        current_timestamp = to_moment(datetime.now(timezone.utc)).timestamp() * 1000

        current_prices: dict[str, float] = {}
        recent_aggregate_trades = slice_deque(
            team.collector.aggregate_trades, 2 ** (10 + 6)
        )
        for symbol in target_symbols:
            for aggregate_trade in reversed(recent_aggregate_trades):
                if aggregate_trade.symbol != symbol:
                    continue
                if aggregate_trade.timestamp < current_timestamp - 60 * 1000:
                    raise ValueError("Recent price is not available for placing orders")
                current_prices[symbol] = aggregate_trade.price
                break

        # ■■■■■ Prepare closure functions ■■■■■

        async def job_cancel_order(payload):
            await self.api_requester.binance(
                http_method="DELETE",
                path="/fapi/v1/allOpenOrders",
                payload=payload,
            )

        async def job_new_order(payload):
            response = await self.api_requester.binance(
                http_method="POST",
                path="/fapi/v1/order",
                payload=payload,
            )
            order_symbol = response["symbol"]
            order_id = response["orderId"]
            timestamp = response["updateTime"] / 1000
            update_time = datetime.fromtimestamp(timestamp, tz=timezone.utc)
            async with self.auto_order_record.write_lock as cell:
                while update_time in cell.data.index:
                    update_time += timedelta(milliseconds=1)
                cell.data.loc[update_time, "SYMBOL"] = order_symbol
                cell.data.loc[update_time, "ORDER_ID"] = order_id
                if not cell.data.index.is_monotonic_increasing:
                    cell.data = await spawn_blocking(sort_data_frame, cell.data)

        # ■■■■■ Do cancel orders ■■■■■

        # These orders must be executed one after another.
        # For example, some `later_orders` expect a position made from `now_orders`
        cancel_orders: list[dict[str, Any]] = []

        for symbol in target_symbols:
            if symbol not in decisions:
                continue

            if OrderType.CANCEL_ALL in decisions[symbol]:
                cancel_order: dict[str, Any] = {
                    "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
                    "symbol": symbol,
                }
                cancel_orders.append(cancel_order)

        if cancel_orders:
            tasks = [spawn(job_cancel_order(o)) for o in cancel_orders]
            await wait(tasks)

        # ■■■■■ Do now orders ■■■■■

        now_orders: list[dict[str, Any]] = []

        for symbol in target_symbols:
            if symbol not in decisions:
                continue

            current_price = current_prices[symbol]
            leverage = self.leverages[symbol]
            maximum_quantity = self.maximum_quantities[symbol]
            minimum_notional = self.minimum_notionals[symbol]
            quantity_precision = self.quantity_precisions[symbol]
            current_direction = self.account_state.positions[symbol].direction

            if OrderType.NOW_CLOSE in decisions[symbol]:
                quantity = maximum_quantity
                if current_direction != PositionDirection.NONE:
                    order_side = (
                        "SELL" if current_direction == PositionDirection.LONG else "BUY"
                    )
                    new_order: dict[str, Any] = {
                        "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
                        "symbol": symbol,
                        "type": "MARKET",
                        "side": order_side,
                        "quantity": quantity,
                        "reduceOnly": True,
                        "newOrderRespType": "RESULT",
                    }
                    now_orders.append(new_order)
                else:
                    text = "Cannot close position when there isn't any"
                    logger.warning(text)

            if OrderType.NOW_BUY in decisions[symbol]:
                decision = decisions[symbol][OrderType.NOW_BUY]
                notional = max(minimum_notional, decision.margin * leverage)
                quantity = min(maximum_quantity, notional / current_price)
                new_order: dict[str, Any] = {
                    "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
                    "symbol": symbol,
                    "type": "MARKET",
                    "side": "BUY",
                    "quantity": ball_ceil(quantity, quantity_precision),
                    "newOrderRespType": "RESULT",
                }
                now_orders.append(new_order)

            if OrderType.NOW_SELL in decisions[symbol]:
                decision = decisions[symbol][OrderType.NOW_SELL]
                notional = max(minimum_notional, decision.margin * leverage)
                quantity = min(maximum_quantity, notional / current_price)
                new_order: dict[str, Any] = {
                    "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
                    "symbol": symbol,
                    "type": "MARKET",
                    "side": "SELL",
                    "quantity": ball_ceil(quantity, quantity_precision),
                    "newOrderRespType": "RESULT",
                }
                now_orders.append(new_order)

        if now_orders:
            tasks = [spawn(job_new_order(o)) for o in now_orders]
            await wait(tasks)

        # ■■■■■ Do book orders ■■■■■

        book_orders: list[dict[str, Any]] = []

        for symbol in target_symbols:
            if symbol not in decisions:
                continue

            leverage = self.leverages[symbol]
            maximum_quantity = self.maximum_quantities[symbol]
            minimum_notional = self.minimum_notionals[symbol]
            price_precision = self.price_precisions[symbol]
            quantity_precision = self.quantity_precisions[symbol]

            if OrderType.BOOK_BUY in decisions[symbol]:
                decision = decisions[symbol][OrderType.BOOK_BUY]
                notional = max(minimum_notional, decision.margin * leverage)
                boundary = decision.boundary
                quantity = min(maximum_quantity, notional / boundary)
                new_order: dict[str, Any] = {
                    "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
                    "symbol": symbol,
                    "type": "LIMIT",
                    "side": "BUY",
                    "quantity": ball_ceil(quantity, quantity_precision),
                    "price": round(boundary, price_precision),
                    "timeInForce": "GTC",
                }
                book_orders.append(new_order)

            if OrderType.BOOK_SELL in decisions[symbol]:
                decision = decisions[symbol][OrderType.BOOK_SELL]
                notional = max(minimum_notional, decision.margin * leverage)
                boundary = decision.boundary
                quantity = min(maximum_quantity, notional / boundary)
                new_order: dict[str, Any] = {
                    "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
                    "symbol": symbol,
                    "type": "LIMIT",
                    "side": "SELL",
                    "quantity": ball_ceil(quantity, quantity_precision),
                    "price": round(boundary, price_precision),
                    "timeInForce": "GTC",
                }
                book_orders.append(new_order)

        if book_orders:
            tasks = [spawn(job_new_order(o)) for o in book_orders]
            await wait(tasks)

        # ■■■■■ Do later orders ■■■■■

        later_orders: list[dict[str, Any]] = []

        for symbol in target_symbols:
            if symbol not in decisions:
                continue

            leverage = self.leverages[symbol]
            maximum_quantity = self.maximum_quantities[symbol]
            minimum_notional = self.minimum_notionals[symbol]
            price_precision = self.price_precisions[symbol]
            quantity_precision = self.quantity_precisions[symbol]
            current_direction = self.account_state.positions[symbol].direction

            # Even if there's no open position analyzed yet
            # due to user data stream from Binance being slow,
            # it's possible to assume that a position would have already been created
            # if there was a `now_buy` or `now_sell` order.
            if current_direction == PositionDirection.NONE:
                if OrderType.NOW_BUY in decisions[symbol]:
                    current_direction = PositionDirection.LONG
                elif OrderType.NOW_SELL in decisions[symbol]:
                    current_direction = PositionDirection.SHORT

            # Even if there's open position analyzed,
            # it's possible to assume that a position would have already been closed
            # if there was a `now_close` order.
            if current_direction != PositionDirection.NONE:
                if OrderType.NOW_CLOSE in decisions[symbol]:
                    current_direction = PositionDirection.NONE

            if OrderType.LATER_UP_CLOSE in decisions[symbol]:
                decision = decisions[symbol][OrderType.LATER_UP_CLOSE]
                if current_direction != PositionDirection.NONE:
                    if current_direction == PositionDirection.LONG:
                        order_side = "SELL"
                        order_type = "TAKE_PROFIT_MARKET"
                    else:
                        order_side = "BUY"
                        order_type = "STOP_MARKET"
                    new_order: dict[str, Any] = {
                        "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
                        "symbol": symbol,
                        "type": order_type,
                        "side": order_side,
                        "stopPrice": round(decision.boundary, price_precision),
                        "closePosition": True,
                    }
                    later_orders.append(new_order)
                else:
                    text = "Cannot place `later_up_close` with no open position"
                    logger.warning(text)

            if OrderType.LATER_DOWN_CLOSE in decisions[symbol]:
                decision = decisions[symbol][OrderType.LATER_DOWN_CLOSE]
                if current_direction != PositionDirection.NONE:
                    if current_direction == PositionDirection.LONG:
                        order_side = "SELL"
                        order_type = "STOP_MARKET"
                    else:
                        order_side = "BUY"
                        order_type = "TAKE_PROFIT_MARKET"
                    new_order: dict[str, Any] = {
                        "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
                        "symbol": symbol,
                        "type": order_type,
                        "side": order_side,
                        "stopPrice": round(decision.boundary, price_precision),
                        "closePosition": True,
                    }
                    later_orders.append(new_order)
                else:
                    text = "Cannot place `later_down_close` with no open position"
                    logger.warning(text)

            if OrderType.LATER_UP_BUY in decisions[symbol]:
                decision = decisions[symbol][OrderType.LATER_UP_BUY]
                notional = max(minimum_notional, decision.margin * leverage)
                boundary = decision.boundary
                quantity = min(maximum_quantity, notional / boundary)
                new_order: dict[str, Any] = {
                    "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
                    "symbol": symbol,
                    "type": "STOP_MARKET",
                    "side": "BUY",
                    "quantity": ball_ceil(quantity, quantity_precision),
                    "stopPrice": round(boundary, price_precision),
                }
                later_orders.append(new_order)

            if OrderType.LATER_DOWN_BUY in decisions[symbol]:
                decision = decisions[symbol][OrderType.LATER_DOWN_BUY]
                notional = max(minimum_notional, decision.margin * leverage)
                boundary = decision.boundary
                quantity = min(maximum_quantity, notional / boundary)
                new_order: dict[str, Any] = {
                    "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
                    "symbol": symbol,
                    "type": "TAKE_PROFIT_MARKET",
                    "side": "BUY",
                    "quantity": ball_ceil(quantity, quantity_precision),
                    "stopPrice": round(boundary, price_precision),
                }
                later_orders.append(new_order)

            if OrderType.LATER_UP_SELL in decisions[symbol]:
                decision = decisions[symbol][OrderType.LATER_UP_SELL]
                notional = max(minimum_notional, decision.margin * leverage)
                boundary = decision.boundary
                quantity = min(maximum_quantity, notional / boundary)
                new_order: dict[str, Any] = {
                    "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
                    "symbol": symbol,
                    "type": "TAKE_PROFIT_MARKET",
                    "side": "SELL",
                    "quantity": ball_ceil(quantity, quantity_precision),
                    "stopPrice": round(boundary, price_precision),
                }
                later_orders.append(new_order)

            if OrderType.LATER_DOWN_SELL in decisions[symbol]:
                decision = decisions[symbol][OrderType.LATER_DOWN_SELL]
                notional = max(minimum_notional, decision.margin * leverage)
                boundary = decision.boundary
                quantity = min(maximum_quantity, notional / boundary)
                new_order: dict[str, Any] = {
                    "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
                    "symbol": symbol,
                    "type": "STOP_MARKET",
                    "side": "SELL",
                    "quantity": ball_ceil(quantity, quantity_precision),
                    "stopPrice": round(boundary, price_precision),
                }
                later_orders.append(new_order)

        if later_orders:
            tasks = [spawn(job_new_order(o)) for o in later_orders]
            await wait(tasks)

    async def clear_positions_and_open_orders(self):
        decisions: dict[str, dict[OrderType, Decision]] = {}
        for symbol in self.window.data_settings.target_symbols:
            decisions[symbol] = {
                OrderType.CANCEL_ALL: Decision(),
                OrderType.NOW_CLOSE: Decision(),
            }
        await self.place_orders(decisions)

    async def cancel_conflicting_orders(self):
        if not self.transaction_settings.should_transact:
            return

        conflicting_order_tuples: list[tuple[str, int]] = []
        for symbol in self.window.data_settings.target_symbols:
            symbol_open_orders = self.account_state.open_orders[symbol]
            grouped_open_orders: dict[OrderType, list[int]] = {}
            for order_id, open_order_state in symbol_open_orders.items():
                order_type = open_order_state.order_type
                if order_type not in grouped_open_orders.keys():
                    grouped_open_orders[order_type] = [order_id]
                else:
                    grouped_open_orders[order_type].append(order_id)
            for order_type, group in grouped_open_orders.items():
                if order_type == OrderType.OTHER:
                    for order_id in group:
                        conflicting_order_tuples.append((symbol, order_id))
                elif len(group) > 1:
                    latest_id = max(group)
                    for order_id in group:
                        if order_id != latest_id:
                            conflicting_order_tuples.append((symbol, order_id))

        async def job(conflicting_order_tuple):
            try:
                payload = {
                    "timestamp": int(datetime.now(timezone.utc).timestamp() * 1000),
                    "symbol": conflicting_order_tuple[0],
                    "orderId": conflicting_order_tuple[1],
                }
                await self.api_requester.binance(
                    http_method="DELETE",
                    path="/fapi/v1/order",
                    payload=payload,
                )
            except ApiRequestError:
                pass

        if conflicting_order_tuples:
            tasks = [spawn(job(c)) for c in conflicting_order_tuples]
            await wait(tasks)

    async def pan_view_range(self):
        if not self.should_draw_frequently:
            return

        widget = self.window.transaction_graph.price_widget
        before_range = widget.getAxis("bottom").range
        range_start = before_range[0]
        range_end = before_range[1]

        if range_end - range_start < 6 * 60 * 60:  # six hours
            return

        widget.setXRange(range_start + 10, range_end + 10, padding=0)  # type:ignore

    async def show_raw_account_state_object(self):
        text = ""

        now_time = datetime.now(timezone.utc)
        time_text = now_time.strftime("%Y-%m-%d %H:%M:%S")
        text += f"At UTC {time_text}"

        text += "\n\n"
        text += self.account_state.model_dump_json(indent=2)

        await overlay(
            "This is the raw account state object",
            LongTextView(text),
        )



from PySide6.QtCore import Qt
from PySide6.QtWidgets import QLabel, QVBoxLayout

from solie.widget import BaseOverlay


class LongTextView(BaseOverlay):
    def __init__(self, long_text: str):
        # ■■■■■ the basic ■■■■■

        super().__init__()

        # ■■■■■ full layout ■■■■■

        full_layout = QVBoxLayout(self)
        cards_layout = QVBoxLayout()
        full_layout.addLayout(cards_layout)

        label = QLabel(long_text)
        label.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        cards_layout.addWidget(label)



import re
from asyncio import sleep
from datetime import datetime, timedelta
from typing import NamedTuple

import numpy as np
import pandas as pd
from pyqtgraph import (
    AxisItem,
    PlotDataItem,
    PlotItem,
    PlotWidget,
    mkBrush,
    mkPen,
)
from PySide6.QtGui import QFont

from solie.utility import (
    AggregateTrade,
    BookTicker,
    MarkPrice,
    PercentAxisItem,
    TimeAxisItem,
)


class LinePair(NamedTuple):
    line_a: PlotDataItem
    line_b: PlotDataItem


class GraphLines:
    def __init__(self):
        # Create widgets.
        self.price_widget = PlotWidget()
        self.volume_widget = PlotWidget()
        self.abstract_widget = PlotWidget()
        self.asset_widget = PlotWidget()

        # Create plots.
        price_plot = self.price_widget.plotItem
        volume_plot = self.volume_widget.plotItem
        abstract_plot = self.abstract_widget.plotItem
        asset_plot = self.asset_widget.plotItem
        if not isinstance(price_plot, PlotItem):
            raise ValueError("Plot item is invalid")
        if not isinstance(volume_plot, PlotItem):
            raise ValueError("Plot item is invalid")
        if not isinstance(abstract_plot, PlotItem):
            raise ValueError("Plot item is invalid")
        if not isinstance(asset_plot, PlotItem):
            raise ValueError("Plot item is invalid")
        self.price_plot = price_plot
        self.volume_plot = volume_plot
        self.abstract_plot = abstract_plot
        self.asset_plot = asset_plot

        # Insert lines in price plots.
        self.book_tickers = LinePair(
            price_plot.plot(
                pen=mkPen("#3F3F3F"),
                connect="finite",
                stepMode="right",
            ),
            price_plot.plot(
                pen=mkPen("#3F3F3F"),
                connect="finite",
                stepMode="right",
            ),
        )
        self.last_price = price_plot.plot(
            pen=mkPen("#5A8CC2"),
            connect="finite",
            stepMode="right",
        )
        self.mark_price = price_plot.plot(
            pen=mkPen("#3E628A"),
            connect="finite",
        )
        self.price_indicators = [price_plot.plot(connect="finite") for _ in range(20)]
        self.entry_price = price_plot.plot(
            pen=mkPen("#FFBB00"),
            connect="finite",
        )
        self.wobbles = LinePair(
            price_plot.plot(
                pen=mkPen("#888888"),
                connect="finite",
                stepMode="right",
            ),
            price_plot.plot(
                pen=mkPen("#888888"),
                connect="finite",
                stepMode="right",
            ),
        )
        self.price_rise = price_plot.plot(
            pen=mkPen("#70E161"),
            connect="finite",
        )
        self.price_fall = price_plot.plot(
            pen=mkPen("#FF304F"),
            connect="finite",
        )
        self.price_stay = price_plot.plot(
            pen=mkPen("#DDDDDD"),
            connect="finite",
        )
        self.sell = price_plot.plot(
            pen=mkPen(None),  # invisible line
            symbol="o",
            symbolBrush="#0055FF",
            symbolPen=mkPen("#BBBBBB"),
            symbolSize=8,
        )
        self.buy = price_plot.plot(
            pen=mkPen(None),  # invisible line
            symbol="o",
            symbolBrush="#FF3300",
            symbolPen=mkPen("#BBBBBB"),
            symbolSize=8,
        )

        # Insert lines in volume plots.
        self.volume = volume_plot.plot(
            pen=mkPen("#BBBBBB"),
            connect="all",
            stepMode="right",
            fillLevel=0,
            brush=mkBrush(255, 255, 255, 15),
        )
        self.last_volume = volume_plot.plot(
            pen=mkPen("#BBBBBB"),
            connect="finite",
        )
        self.volume_indicators = [volume_plot.plot(connect="finite") for _ in range(20)]

        # Insert lines in abstract plots.
        self.abstract_indicators = [
            abstract_plot.plot(connect="finite") for _ in range(20)
        ]

        # Insert lines in asset plots.
        self.asset_with_unrealized_profit = asset_plot.plot(
            pen=mkPen("#999999"),
            connect="finite",
        )
        self.asset = asset_plot.plot(
            pen=mkPen("#FF8700"),
            connect="finite",
            stepMode="right",
        )

        # Configure UX.
        self._configure_widgets()
        self._configure_plots()

    def _configure_widgets(self):
        self.price_widget.setBackground("#252525")
        self.volume_widget.setBackground("#252525")
        self.abstract_widget.setBackground("#252525")
        self.asset_widget.setBackground("#252525")

        self.price_widget.setMouseEnabled(y=False)
        self.volume_widget.setMouseEnabled(y=False)
        self.abstract_widget.setMouseEnabled(y=False)
        self.asset_widget.setMouseEnabled(y=False)

        self.price_widget.enableAutoRange(y=True)
        self.volume_widget.enableAutoRange(y=True)
        self.abstract_widget.enableAutoRange(y=True)
        self.asset_widget.enableAutoRange(y=True)

        self.volume_widget.setXLink(self.price_widget)
        self.abstract_widget.setXLink(self.volume_widget)
        self.asset_widget.setXLink(self.abstract_widget)

    def _configure_plots(self):
        self.price_plot.vb.setLimits(xMin=0, yMin=0)  # type:ignore
        self.volume_plot.vb.setLimits(xMin=0, yMin=0)  # type:ignore
        self.abstract_plot.vb.setLimits(xMin=0)  # type:ignore
        self.asset_plot.vb.setLimits(xMin=0, yMin=0)  # type:ignore

        self.price_plot.setDownsampling(auto=True, mode="subsample")
        self.price_plot.setClipToView(True)
        self.price_plot.setAutoVisible(y=True)  # type:ignore
        self.volume_plot.setDownsampling(auto=True, mode="subsample")
        self.volume_plot.setClipToView(True)
        self.volume_plot.setAutoVisible(y=True)  # type:ignore
        self.abstract_plot.setDownsampling(auto=True, mode="subsample")
        self.abstract_plot.setClipToView(True)
        self.abstract_plot.setAutoVisible(y=True)  # type:ignore
        self.asset_plot.setDownsampling(auto=True, mode="subsample")
        self.asset_plot.setClipToView(True)
        self.asset_plot.setAutoVisible(y=True)  # type:ignore

        axis_items = {
            "top": TimeAxisItem(orientation="top"),
            "bottom": TimeAxisItem(orientation="bottom"),
            "left": PercentAxisItem(orientation="left"),
            "right": PercentAxisItem(orientation="right"),
        }
        self.price_plot.setAxisItems(axis_items)
        axis_items = {
            "top": TimeAxisItem(orientation="top"),
            "bottom": TimeAxisItem(orientation="bottom"),
            "left": AxisItem(orientation="left"),
            "right": AxisItem(orientation="right"),
        }
        self.volume_plot.setAxisItems(axis_items)
        axis_items = {
            "top": TimeAxisItem(orientation="top"),
            "bottom": TimeAxisItem(orientation="bottom"),
            "left": AxisItem(orientation="left"),
            "right": AxisItem(orientation="right"),
        }
        self.abstract_plot.setAxisItems(axis_items)
        axis_items = {
            "top": TimeAxisItem(orientation="top"),
            "bottom": TimeAxisItem(orientation="bottom"),
            "left": PercentAxisItem(orientation="left"),
            "right": PercentAxisItem(orientation="right"),
        }
        self.asset_plot.setAxisItems(axis_items)

        tick_font = QFont("Source Code Pro", 7)
        self.price_plot.getAxis("top").setTickFont(tick_font)
        self.price_plot.getAxis("bottom").setTickFont(tick_font)
        self.price_plot.getAxis("left").setTickFont(tick_font)
        self.price_plot.getAxis("right").setTickFont(tick_font)
        self.volume_plot.getAxis("top").setTickFont(tick_font)
        self.volume_plot.getAxis("bottom").setTickFont(tick_font)
        self.volume_plot.getAxis("left").setTickFont(tick_font)
        self.volume_plot.getAxis("right").setTickFont(tick_font)
        self.abstract_plot.getAxis("top").setTickFont(tick_font)
        self.abstract_plot.getAxis("bottom").setTickFont(tick_font)
        self.abstract_plot.getAxis("left").setTickFont(tick_font)
        self.abstract_plot.getAxis("right").setTickFont(tick_font)
        self.asset_plot.getAxis("top").setTickFont(tick_font)
        self.asset_plot.getAxis("bottom").setTickFont(tick_font)
        self.asset_plot.getAxis("left").setTickFont(tick_font)
        self.asset_plot.getAxis("right").setTickFont(tick_font)

        self.price_plot.getAxis("left").setWidth(40)
        self.price_plot.getAxis("right").setWidth(40)
        self.volume_plot.getAxis("left").setWidth(40)
        self.volume_plot.getAxis("right").setWidth(40)
        self.abstract_plot.getAxis("left").setWidth(40)
        self.abstract_plot.getAxis("right").setWidth(40)
        self.asset_plot.getAxis("left").setWidth(40)
        self.asset_plot.getAxis("right").setWidth(40)

        self.price_plot.getAxis("bottom").setHeight(0)
        self.volume_plot.getAxis("top").setHeight(0)
        self.volume_plot.getAxis("bottom").setHeight(0)
        self.abstract_plot.getAxis("top").setHeight(0)
        self.abstract_plot.getAxis("bottom").setHeight(0)
        self.asset_plot.getAxis("top").setHeight(0)

        self.price_plot.showGrid(x=True, y=True, alpha=0.1)
        self.volume_plot.showGrid(x=True, y=True, alpha=0.1)
        self.abstract_plot.showGrid(x=True, y=True, alpha=0.1)
        self.asset_plot.showGrid(x=True, y=True, alpha=0.1)

    async def update_light_lines(
        self,
        mark_prices: list[MarkPrice],
        aggregate_trades: list[AggregateTrade],
        book_tickers: list[BookTicker],
        entry_price: float | None,
        observed_until: datetime,
    ):
        # mark price
        data_y = [d.mark_price for d in mark_prices]
        data_x = [d.timestamp / 10**3 for d in mark_prices]
        self.mark_price.setData(data_x, data_y)
        await sleep(0.0)

        # last price
        timestamps = [t.timestamp / 10**3 for t in aggregate_trades]

        data_x = timestamps.copy()
        data_y = [t.price for t in aggregate_trades]
        self.last_price.setData(data_x, data_y)
        await sleep(0.0)

        # last trade volume
        index_ar = np.array(timestamps)
        value_ar = np.array([t.volume for t in aggregate_trades])
        length = len(index_ar)
        zero_ar = np.zeros(length)
        nan_ar = np.empty(length)
        nan_ar[:] = np.nan
        data_x = np.repeat(index_ar, 3)
        data_y = np.stack([nan_ar, zero_ar, value_ar], axis=1).reshape(-1)
        self.last_volume.setData(data_x, data_y)
        await sleep(0.0)

        # book tickers
        data_x = [d.timestamp / 10**3 for d in book_tickers]
        data_y = [d.best_bid_price for d in book_tickers]
        self.book_tickers.line_a.setData(data_x, data_y)
        data_y = [d.best_ask_price for d in book_tickers]
        self.book_tickers.line_b.setData(data_x, data_y)
        await sleep(0.0)

        # entry price
        first_moment = observed_until - timedelta(hours=12)
        last_moment = observed_until + timedelta(hours=12)
        if entry_price is None:
            data_x = []
            data_y = []
        else:
            data_x = np.linspace(
                first_moment.timestamp(),
                last_moment.timestamp(),
                num=1000,
            )
            data_y = np.linspace(
                entry_price,
                entry_price,
                num=1000,
            )
        self.entry_price.setData(data_x, data_y)
        await sleep(0.0)

    async def update_heavy_lines(
        self,
        symbol: str,
        candle_data: pd.DataFrame,
        asset_record: pd.DataFrame,
        unrealized_changes: pd.Series,
    ):
        # price movement
        index_ar = candle_data.index.to_numpy(dtype=np.int64) / 10**9
        open_ar = candle_data[f"{symbol}/OPEN"].to_numpy()
        close_ar = candle_data[f"{symbol}/CLOSE"].to_numpy()
        high_ar = candle_data[f"{symbol}/HIGH"].to_numpy()
        low_ar = candle_data[f"{symbol}/LOW"].to_numpy()
        rise_ar = close_ar > open_ar
        fall_ar = close_ar < open_ar
        stay_ar = close_ar == open_ar
        length = len(index_ar)
        nan_ar = np.empty(length)
        nan_ar[:] = np.nan

        data_x = np.stack(
            [
                index_ar[rise_ar] + 2,
                index_ar[rise_ar] + 5,
                index_ar[rise_ar],
                index_ar[rise_ar] + 5,
                index_ar[rise_ar] + 8,
                index_ar[rise_ar],
                index_ar[rise_ar] + 5,
                index_ar[rise_ar] + 5,
                index_ar[rise_ar],
            ],
            axis=1,
        ).reshape(-1)
        data_y = np.stack(
            [
                open_ar[rise_ar],
                open_ar[rise_ar],
                nan_ar[rise_ar],
                close_ar[rise_ar],
                close_ar[rise_ar],
                nan_ar[rise_ar],
                high_ar[rise_ar],
                low_ar[rise_ar],
                nan_ar[rise_ar],
            ],
            axis=1,
        ).reshape(-1)
        self.price_rise.setData(data_x, data_y)
        await sleep(0.0)

        data_x = np.stack(
            [
                index_ar[fall_ar] + 2,
                index_ar[fall_ar] + 5,
                index_ar[fall_ar],
                index_ar[fall_ar] + 5,
                index_ar[fall_ar] + 8,
                index_ar[fall_ar],
                index_ar[fall_ar] + 5,
                index_ar[fall_ar] + 5,
                index_ar[fall_ar],
            ],
            axis=1,
        ).reshape(-1)
        data_y = np.stack(
            [
                open_ar[fall_ar],
                open_ar[fall_ar],
                nan_ar[fall_ar],
                close_ar[fall_ar],
                close_ar[fall_ar],
                nan_ar[fall_ar],
                high_ar[fall_ar],
                low_ar[fall_ar],
                nan_ar[fall_ar],
            ],
            axis=1,
        ).reshape(-1)
        self.price_fall.setData(data_x, data_y)
        await sleep(0.0)

        data_x = np.stack(
            [
                index_ar[stay_ar] + 2,
                index_ar[stay_ar] + 5,
                index_ar[stay_ar],
                index_ar[stay_ar] + 5,
                index_ar[stay_ar] + 8,
                index_ar[stay_ar],
                index_ar[stay_ar] + 5,
                index_ar[stay_ar] + 5,
                index_ar[stay_ar],
            ],
            axis=1,
        ).reshape(-1)
        data_y = np.stack(
            [
                open_ar[stay_ar],
                open_ar[stay_ar],
                nan_ar[stay_ar],
                close_ar[stay_ar],
                close_ar[stay_ar],
                nan_ar[stay_ar],
                high_ar[stay_ar],
                low_ar[stay_ar],
                nan_ar[stay_ar],
            ],
            axis=1,
        ).reshape(-1)
        self.price_stay.setData(data_x, data_y)
        await sleep(0.0)

        # wobbles
        sr = candle_data[f"{symbol}/HIGH"]
        data_x = sr.index.to_numpy(dtype=np.int64) / 10**9
        data_y = sr.to_numpy(dtype=np.float32)
        self.wobbles.line_a.setData(data_x, data_y)
        await sleep(0.0)

        sr = candle_data[f"{symbol}/LOW"]
        data_x = sr.index.to_numpy(dtype=np.int64) / 10**9
        data_y = sr.to_numpy(dtype=np.float32)
        self.wobbles.line_b.setData(data_x, data_y)
        await sleep(0.0)

        # trade volume
        sr = candle_data[f"{symbol}/VOLUME"]
        sr = sr.fillna(value=0)
        data_x = sr.index.to_numpy(dtype=np.int64) / 10**9
        data_y = sr.to_numpy(dtype=np.float32)
        self.volume.setData(data_x, data_y)
        await sleep(0.0)

        # asset
        data_x = asset_record["RESULT_ASSET"].index.to_numpy(dtype=np.int64) / 10**9
        data_y = asset_record["RESULT_ASSET"].to_numpy(dtype=np.float32)
        self.asset.setData(data_x, data_y)
        await sleep(0.0)

        # asset with unrealized profit
        sr = asset_record["RESULT_ASSET"]
        if len(sr) >= 2:
            sr = sr.resample("10s").ffill()
        unrealized_changes_sr = unrealized_changes.reindex(sr.index)
        sr = sr * (1 + unrealized_changes_sr)
        data_x = sr.index.to_numpy(dtype=np.int64) / 10**9 + 5
        data_y = sr.to_numpy(dtype=np.float32)
        self.asset_with_unrealized_profit.setData(data_x, data_y)
        await sleep(0.0)

        # buy and sell
        df = asset_record.loc[asset_record["SYMBOL"] == symbol]
        df = df[df["SIDE"] == "SELL"]
        sr = df["FILL_PRICE"]
        data_x = sr.index.to_numpy(dtype=np.int64) / 10**9
        data_y = sr.to_numpy(dtype=np.float32)
        self.sell.setData(data_x, data_y)
        await sleep(0.0)

        df = asset_record.loc[asset_record["SYMBOL"] == symbol]
        df = df[df["SIDE"] == "BUY"]
        sr = df["FILL_PRICE"]
        data_x = sr.index.to_numpy(dtype=np.int64) / 10**9
        data_y = sr.to_numpy(dtype=np.float32)
        self.buy.setData(data_x, data_y)
        await sleep(0.0)

    async def update_custom_lines(self, symbol: str, indicators: pd.DataFrame):
        # common data
        columns = [str(n) for n in indicators.columns]
        data_x = indicators.index.to_numpy(dtype=np.int64) / 10**9
        data_x += 5

        # price indicators
        chosen_columns = [n for n in columns if n.startswith(f"{symbol}/PRICE")]
        for index, widget in enumerate(self.price_indicators):
            if index >= len(chosen_columns):
                widget.clear()
                continue
            column = chosen_columns[index]
            sr = indicators[column]
            data_y = sr.to_numpy(dtype=np.float32)
            inside_strings = re.findall(r"\(([^)]+)", column)
            if len(inside_strings) == 0:
                color = "#AAAAAA"
            else:
                color = inside_strings[0]
            widget.setPen(color)
            widget.setData(data_x, data_y)
            await sleep(0.0)

        # trade volume indicators
        chosen_columns = [n for n in columns if n.startswith(f"{symbol}/VOLUME")]
        for index, widget in enumerate(self.volume_indicators):
            if index >= len(chosen_columns):
                widget.clear()
                continue
            column = chosen_columns[index]
            sr = indicators[column]
            data_y = sr.to_numpy(dtype=np.float32)
            inside_strings = re.findall(r"\(([^)]+)", column)
            if len(inside_strings) == 0:
                color = "#AAAAAA"
            else:
                color = inside_strings[0]
            widget.setPen(color)
            widget.setData(data_x, data_y)
            await sleep(0.0)

        # abstract indicators
        chosen_columns = [n for n in columns if n.startswith(f"{symbol}/ABSTRACT")]
        for index, widget in enumerate(self.abstract_indicators):
            if index >= len(chosen_columns):
                widget.clear()
                continue
            column = chosen_columns[index]
            sr = indicators[column]
            data_y = sr.to_numpy(dtype=np.float32)
            inside_strings = re.findall(r"\(([^)]+)", column)
            if len(inside_strings) == 0:
                color = "#AAAAAA"
            else:
                color = inside_strings[0]
            widget.setPen(color)
            widget.setData(data_x, data_y)
            await sleep(0.0)



from pygments import lex
from pygments.lexers import get_lexer_by_name
from PySide6.QtGui import QBrush, QColor, QSyntaxHighlighter, QTextCharFormat
from PySide6.QtWidgets import QWidget
from typing_extensions import override


class SyntaxHighlighter(QSyntaxHighlighter):
    @override
    def highlightBlock(self, text: str):
        current_position = 0

        for token, text_block in lex(text, self._lexer):
            length = len(text_block)
            text_format = None
            for text_type in self._mapping.keys():
                if str(token).startswith(text_type):
                    text_format = self._mapping[text_type]
                    break
            if text_format is not None:
                self.setFormat(current_position, length, text_format)
            current_position += length

    def __init__(self, parent: QWidget):
        super().__init__(parent)

        self._lexer = get_lexer_by_name("python")
        self._mapping: dict[str, QTextCharFormat] = {}

        text_type = "Token.Keyword"
        hex_color = "#d1939e"
        text_format = QTextCharFormat()
        text_format.setForeground(QBrush(QColor(hex_color)))
        self._mapping[text_type] = text_format

        text_type = "Token.Literal.String"
        hex_color = "#bce051"
        text_format = QTextCharFormat()
        text_format.setForeground(QBrush(QColor(hex_color)))
        self._mapping[text_type] = text_format

        text_type = "Token.Literal.Number"
        hex_color = "#d1939e"
        text_format = QTextCharFormat()
        text_format.setForeground(QBrush(QColor(hex_color)))
        self._mapping[text_type] = text_format

        text_type = "Token.Operator"
        hex_color = "#f4b73d"
        text_format = QTextCharFormat()
        text_format.setForeground(QBrush(QColor(hex_color)))
        self._mapping[text_type] = text_format

        text_type = "Token.Punctuation"
        hex_color = "#999999"
        text_format = QTextCharFormat()
        text_format.setForeground(QBrush(QColor(hex_color)))
        self._mapping[text_type] = text_format

        text_type = "Token.Comment"
        hex_color = "#997f66"
        text_format = QTextCharFormat()
        text_format.setForeground(QBrush(QColor(hex_color)))
        self._mapping[text_type] = text_format



from .entry import bring_to_life
from .utility import (
    AccountState,
    Decision,
    DecisionInput,
    IndicatorInput,
    OpenOrder,
    OrderType,
    Position,
    PositionDirection,
    RiskLevel,
    SolieConfig,
    Strategy,
)
from .worker import Team, team

__all__ = [
    "AccountState",
    "Decision",
    "DecisionInput",
    "IndicatorInput",
    "OpenOrder",
    "OrderType",
    "Position",
    "PositionDirection",
    "RiskLevel",
    "SolieConfig",
    "Strategy",
    "Team",
    "bring_to_life",
    "team",
]



from collections.abc import Callable, Coroutine
from typing import Any

from PySide6.QtCore import SignalInstance

from .concurrency import spawn


def outsource(
    signal: SignalInstance, target_function: Callable[[], Coroutine[None, None, Any]]
):
    def job(*args, **kwargs):
        spawn(target_function())

    signal.connect(job)



import sys
from asyncio import run
from typing import NoReturn

from PySide6.QtWidgets import QApplication

from solie.common import prepare_process_pool
from solie.utility import SolieConfig

from .lifetime import live


def bring_to_life(config: SolieConfig | None = None) -> NoReturn:
    prepare_process_pool()

    if config is None:
        config = SolieConfig()

    app = QApplication()
    run(live(app, config))

    sys.exit()



import logging
from asyncio import Event, sleep, wait

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from PySide6.QtGui import QColor, QFont, QFontDatabase, QPalette
from PySide6.QtWidgets import QApplication

from solie.common import PACKAGE_NAME, PACKAGE_PATH, spawn
from solie.utility import SolieConfig
from solie.widget import AskPopup, OverlayPopup
from solie.window import Window
from solie.worker import (
    Collector,
    Manager,
    Simulator,
    Strategiest,
    Transactor,
    team,
)

logger = logging.getLogger(__name__)


async def keep_processing_events(app: QApplication):
    """
    `Qt` does not have proper async support, as it is focused on threads.
    To use async-based concurrency, we need to periodically process UI events.
    Third-party polling libraries are not very reliable.
    """
    interval = 1 / 240
    while True:
        app.processEvents()
        await sleep(interval)


async def live(app: QApplication, config: SolieConfig):
    staticpath = PACKAGE_PATH / "static"
    QFontDatabase.addApplicationFont(str(staticpath / "source_code_pro.ttf"))
    QFontDatabase.addApplicationFont(str(staticpath / "notosans_regular.ttf"))
    QFontDatabase.addApplicationFont(str(staticpath / "lexend_bold.ttf"))
    default_font = QFont("Noto Sans", 9)
    app.setFont(default_font)

    dark_palette = QPalette()
    dark_palette.setColor(QPalette.ColorRole.Window, QColor(29, 29, 29))
    dark_palette.setColor(QPalette.ColorRole.WindowText, QColor(230, 230, 230))
    dark_palette.setColor(QPalette.ColorRole.Base, QColor(22, 22, 22))
    dark_palette.setColor(QPalette.ColorRole.AlternateBase, QColor(29, 29, 29))
    dark_palette.setColor(QPalette.ColorRole.ToolTipBase, QColor(230, 230, 230))
    dark_palette.setColor(QPalette.ColorRole.ToolTipText, QColor(230, 230, 230))
    dark_palette.setColor(QPalette.ColorRole.Text, QColor(230, 230, 230))
    dark_palette.setColor(QPalette.ColorRole.Button, QColor(29, 29, 29))
    dark_palette.setColor(QPalette.ColorRole.ButtonText, QColor(230, 230, 230))
    dark_palette.setColor(QPalette.ColorRole.BrightText, QColor(255, 180, 0))
    dark_palette.setColor(QPalette.ColorRole.Link, QColor(42, 130, 218))
    dark_palette.setColor(QPalette.ColorRole.Highlight, QColor(42, 130, 218))
    dark_palette.setColor(QPalette.ColorRole.HighlightedText, QColor(0, 0, 0))
    app.setStyle("Fusion")
    app.setPalette(dark_palette)

    close_event = Event()
    scheduler = AsyncIOScheduler(timezone="UTC")

    window = Window(close_event, config)
    window.setPalette(dark_palette)
    AskPopup.install_window(window)
    OverlayPopup.install_window(window)
    spawn(keep_processing_events(app))

    logging.getLogger(PACKAGE_NAME).setLevel("DEBUG")
    await window.boot()
    logger.info("Started up")

    collector = Collector(window, scheduler)
    transactor = Transactor(window, scheduler)
    simulator = Simulator(window, scheduler)
    strategist = Strategiest(window, scheduler)
    manager = Manager(window, scheduler)

    team.unite(collector, transactor, simulator, strategist, manager)

    tasks = [
        spawn(collector.load()),
        spawn(transactor.load()),
        spawn(simulator.load()),
        spawn(strategist.load()),
        spawn(manager.load()),
    ]
    await wait(tasks)

    spawn(collector.get_exchange_information())
    spawn(strategist.display_strategies())
    spawn(transactor.display_strategy_index())
    spawn(transactor.watch_binance())
    spawn(transactor.update_user_data_stream())
    spawn(transactor.display_lines())
    spawn(transactor.display_day_range())
    spawn(simulator.display_lines())
    spawn(simulator.display_year_range())
    spawn(simulator.display_available_years())
    spawn(manager.check_binance_limits())
    spawn(manager.display_internal_status())

    scheduler.start()
    await sleep(1)

    window.reveal()
    await close_event.wait()

    scheduler.shutdown()
    await sleep(1)

    tasks = [
        spawn(transactor.save_large_data()),
        spawn(transactor.save_scribbles()),
        spawn(strategist.save_strategies()),
        spawn(collector.save_candle_data()),
    ]
    await wait(tasks)



import random
from datetime import datetime, timezone
from itertools import product

import numpy as np
import pandas as pd

from .data_models import AccountState, Position, PositionDirection


def create_empty_candle_data(target_symbols: list[str]) -> pd.DataFrame:
    column_pairs = product(
        target_symbols,
        ("OPEN", "HIGH", "LOW", "CLOSE", "VOLUME"),
    )
    return pd.DataFrame(
        columns=["/".join(p) for p in column_pairs],
        dtype=np.float32,
        index=pd.DatetimeIndex([], tz="UTC"),
    )


def create_empty_account_state(target_symbols: list[str]) -> AccountState:
    return AccountState(
        observed_until=datetime.fromtimestamp(0.0, tz=timezone.utc),
        wallet_balance=1.0,
        positions={
            s: Position(
                margin=0.0,
                direction=PositionDirection.NONE,
                entry_price=0.0,
                update_time=datetime.fromtimestamp(0.0, tz=timezone.utc),
            )
            for s in target_symbols
        },
        open_orders={s: {} for s in target_symbols},
    )


def create_empty_asset_record():
    return pd.DataFrame(
        columns=[
            "CAUSE",
            "SYMBOL",
            "SIDE",
            "FILL_PRICE",
            "ROLE",
            "MARGIN_RATIO",
            "ORDER_ID",
            "RESULT_ASSET",
        ],
        index=pd.DatetimeIndex([], tz="UTC"),
    )


def create_empty_unrealized_changes() -> pd.Series:
    return pd.Series(index=pd.DatetimeIndex([], tz="UTC"), dtype=np.float32)


def create_strategy_code_name() -> str:
    ingredients = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    code_name = "".join(random.choice(ingredients) for _ in range(6))
    return code_name



from pathlib import Path
from types import CodeType

import aiofiles
import aiofiles.os
from pydantic import BaseModel

from solie.common import PACKAGE_PATH

from .data_models import (
    DataSettings,
    DecisionInput,
    IndicatorInput,
    Strategy,
)

DATAPATH_FILE = PACKAGE_PATH / "datapath.txt"


async def read_datapath() -> Path | None:
    if await aiofiles.os.path.isfile(DATAPATH_FILE):
        async with aiofiles.open(DATAPATH_FILE, "r", encoding="utf8") as file:
            datapath = Path(await file.read())
        if await aiofiles.os.path.isdir(datapath):
            return datapath
        else:
            return None
    else:
        return None


async def save_datapath(datapath: Path | None):
    if datapath:
        async with aiofiles.open(DATAPATH_FILE, "w", encoding="utf8") as file:
            await file.write(str(datapath))
    else:
        await aiofiles.os.remove(DATAPATH_FILE)


async def read_data_settings(datapath: Path) -> DataSettings | None:
    filepath = datapath / "data_settings.json"
    if await aiofiles.os.path.isfile(filepath):
        async with aiofiles.open(filepath, "r", encoding="utf8") as file:
            data_settings = DataSettings.model_validate_json(await file.read())
        return data_settings
    else:
        return None


async def save_data_settings(data_settings: DataSettings, datapath: Path):
    filepath = datapath / "data_settings.json"
    async with aiofiles.open(filepath, "w", encoding="utf8") as file:
        await file.write(data_settings.model_dump_json(indent=2))


class SavedStrategy(Strategy):
    indicator_script: str = "pass"
    decision_script: str = "pass"

    _compiled_indicator_script: CodeType | None = None
    _compiled_decision_script: CodeType | None = None

    def create_indicators(self, given: IndicatorInput):
        target_symbols = given.target_symbols
        candle_data = given.candle_data
        new_indicators = given.new_indicators

        if self._compiled_indicator_script is None:
            code = self.indicator_script
        else:
            code = self._compiled_indicator_script

        namespace = {
            "target_symbols": target_symbols,
            "candle_data": candle_data,
            "new_indicators": new_indicators,
        }
        exec(code, namespace)

    def create_decisions(self, given: DecisionInput):
        target_symbols = given.target_symbols
        account_state = given.account_state
        current_moment = given.current_moment
        current_candle_data = given.current_candle_data
        current_indicators = given.current_indicators
        scribbles = given.scribbles
        new_decisions = given.new_decisions

        if self._compiled_decision_script is None:
            code = self.decision_script
        else:
            code = self._compiled_decision_script

        namespace = {
            "target_symbols": target_symbols,
            "current_moment": current_moment,
            "current_candle_data": current_candle_data,
            "current_indicators": current_indicators,
            "account_state": account_state,
            "scribbles": scribbles,
            "decisions": new_decisions,
        }
        exec(code, namespace)

    def compile_code(self):
        """
        Enables faster execution of the strategy code
        by precompiling the text-based script.
        This is needed for high-performance simulation.
        After this method is called, the instance becomes unpicklable.
        """
        self._compiled_indicator_script = compile(
            self.indicator_script, "<string>", "exec"
        )
        self._compiled_decision_script = compile(
            self.decision_script, "<string>", "exec"
        )


class SavedStrategies(BaseModel):
    all: list[SavedStrategy]


class SolieConfig:
    def __init__(self):
        self.strategies: list[Strategy] = []

    def add_strategy(self, strategy: Strategy):
        self.strategies.append(strategy)



import logging
import time
from asyncio import Lock
from collections.abc import Callable
from datetime import datetime, timezone
from pathlib import Path

import aiofiles
import aiofiles.os

from solie.common import spawn


class LogHandler(logging.Handler):
    file_lock = Lock()

    def __init__(self, log_path: Path, callback: Callable[[str, str], None]):
        super().__init__()

        self.log_path = log_path
        self.callback = callback

        log_format = "%(asctime)s.%(msecs)03d %(levelname)s"
        date_format = "%Y-%m-%d %H:%M:%S"
        log_formatter = logging.Formatter(log_format, datefmt=date_format)
        log_formatter.converter = time.gmtime
        self.setFormatter(log_formatter)

        now = datetime.now(timezone.utc).replace(microsecond=0)
        self.filename = (
            f"{now.year:04}-{now.month:02}-{now.day:02}"
            + f".{now.hour:02}-{now.minute:02}-{now.second:02}"
            + f".{now.tzinfo}.txt"
        )

    def emit(self, record):
        formatted = self.format(record)

        if record.exc_info is not None:
            # when this is from an exception
            lines = formatted.split("\n")
            summarization = lines[0]
            log_content = "\n".join(lines[1:])
            exception = record.exc_info[0]
            if exception is None:
                return
            summarization += f" - {exception.__name__}"
        else:
            # when this is a normal log
            summarization = formatted
            log_content = record.getMessage()
            first_line_content = log_content.split("\n")[0].strip()
            summarization += f" - {first_line_content}"
            summarization = summarization[:80]

        log_content = f"{formatted}\n{log_content}"

        spawn(self.add_log_output(summarization, log_content))

    async def add_log_output(self, summarization: str, log_content: str):
        # add to log list
        self.callback(summarization, log_content)

        # save to file
        filepath = self.log_path / self.filename
        async with self.file_lock:
            async with aiofiles.open(filepath, "a", encoding="utf8") as file:
                line_divider = "-" * 80
                await file.write(f"{log_content}\n\n{line_divider}\n\n")



from abc import abstractmethod
from datetime import datetime
from enum import Enum
from typing import Any, NamedTuple

import pandas as pd
from pydantic import BaseModel

# We use `BaseModel` when parsing, validation, or mutability is needed.
# Otherwise, `NamedTuple` is preferred because it's more performant.


class DataSettings(BaseModel):
    asset_token: str
    target_symbols: list[str]


class TransactionSettings(BaseModel):
    strategy_index: int = 0
    should_transact: bool = False
    desired_leverage: int = 1
    binance_api_key: str = ""
    binance_api_secret: str = ""


class SimulationSettings(BaseModel):
    year: int
    strategy_index: int = 0
    maker_fee: float = 0.02
    taker_fee: float = 0.04
    leverage: int = 1


class SimulationSummary(BaseModel):
    year: int
    strategy_code_name: str
    strategy_version: str


class BoardLockOptions(Enum):
    NEVER = 0
    SECONDS_10 = 1
    MINUTE_1 = 2
    MINUTE_10 = 3
    HOUR_1 = 4


class ManagementSettings(BaseModel):
    lock_board: BoardLockOptions = BoardLockOptions.NEVER


class BookTicker(NamedTuple):
    timestamp: int  # In milliseconds
    symbol: str
    best_bid_price: float
    best_ask_price: float


class MarkPrice(NamedTuple):
    timestamp: int  # In milliseconds
    symbol: str
    mark_price: float


class AggregateTrade(NamedTuple):
    timestamp: int  # In milliseconds
    symbol: str
    price: float
    volume: float


class OrderType(Enum):
    NOW_BUY = 0
    NOW_SELL = 1
    NOW_CLOSE = 2
    CANCEL_ALL = 3
    BOOK_BUY = 4
    BOOK_SELL = 5
    LATER_UP_BUY = 6
    LATER_UP_SELL = 7
    LATER_UP_CLOSE = 8
    LATER_DOWN_BUY = 9
    LATER_DOWN_SELL = 10
    LATER_DOWN_CLOSE = 11
    OTHER = 12

    def is_now(self) -> bool:
        return self in (
            OrderType.NOW_BUY,
            OrderType.NOW_SELL,
            OrderType.NOW_CLOSE,
        )

    def is_book(self) -> bool:
        return self in (
            OrderType.BOOK_BUY,
            OrderType.BOOK_SELL,
        )

    def is_later(self) -> bool:
        return self in (
            OrderType.LATER_UP_BUY,
            OrderType.LATER_UP_SELL,
            OrderType.LATER_UP_CLOSE,
            OrderType.LATER_DOWN_BUY,
            OrderType.LATER_DOWN_SELL,
            OrderType.LATER_DOWN_CLOSE,
        )


class PositionDirection(Enum):
    SHORT = -1
    NONE = 0
    LONG = 1


class Decision(BaseModel):
    boundary: float = 0.0
    margin: float = 0.0


class OpenOrder(BaseModel):
    order_type: OrderType
    boundary: float
    """The price where this order gains effect"""
    left_margin: float | None
    """Amount of the asset to be invested, in dollars"""


class Position(BaseModel):
    margin: float
    direction: PositionDirection
    entry_price: float
    update_time: datetime
    """Time of the last trade in this position"""


class AccountState(BaseModel):
    observed_until: datetime
    wallet_balance: float
    """Total assets, in dollars"""
    positions: dict[str, Position]
    """Current position direction of a specific symbol"""
    open_orders: dict[str, dict[int, OpenOrder]]


class VirtualPosition(BaseModel):
    """Virtual position inside simulation"""

    amount: float
    entry_price: float


class VirtualPlacement(BaseModel):
    """Virtual order that has been placed during simulation"""

    boundary: float
    """The price where this order gains effect"""
    margin: float
    """Amount of the asset to be invested, in dollars"""
    order_id: int


class VirtualState(BaseModel):
    """Virtual account state used in simulation"""

    available_balance: float
    positions: dict[str, VirtualPosition]
    placements: dict[str, dict[OrderType, VirtualPlacement]]


class RiskLevel(Enum):
    LOW = 0
    MIDDLE = 1
    HIGH = 2


class IndicatorInput(NamedTuple):
    target_symbols: list[str]
    candle_data: pd.DataFrame
    new_indicators: dict[str, pd.Series]


class DecisionInput(NamedTuple):
    target_symbols: list[str]
    account_state: AccountState
    current_moment: datetime
    current_candle_data: dict[str, float]
    current_indicators: dict[str, float]
    scribbles: dict[Any, Any]
    new_decisions: dict[str, dict[OrderType, Decision]]


class Strategy(BaseModel):
    code_name: str
    readable_name: str = "New Blank Strategy"
    version: str = "0.1"
    description: str = "A blank strategy template before being written"
    risk_level: RiskLevel = RiskLevel.HIGH
    parallelized_simulation: bool = False
    chunk_division: int = 30

    @abstractmethod
    def create_indicators(self, given: IndicatorInput):
        pass

    @abstractmethod
    def create_decisions(self, given: DecisionInput):
        pass



from PySide6.QtWidgets import QFrame, QWidget


class HorizontalDivider(QFrame):
    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.setFrameShape(QFrame.Shape.HLine)
        self.setFixedHeight(2)



from asyncio import Task, create_task
from collections.abc import Callable, Coroutine
from typing import Any, TypeVar

T = TypeVar("T")

# A set to keep track of all running tasks.
all_tasks = set[Task[Any]]()


def spawn(coroutine: Coroutine[None, None, T]) -> Task[T]:
    """
    Spawns an asynchronous task from the given coroutine and manages its lifecycle.

    This function creates a new `asyncio.Task` from the provided coroutine and
    adds it to a global set to maintain a strong reference. This
    prevents the task from being prematurely garbage-collected by the event loop.

    Once the task completes, it automatically removes itself from the set to
    avoid memory leaks.
    """

    task = create_task(coroutine)

    # Add task to the set. This creates a strong reference.
    # Per the `asyncio` documentation,
    # the event loop only retains a weak reference to tasks.
    # If the task returned by `asyncio.create_task` and
    # `asyncio.ensure_future` is not stored in
    # a variable, or a collection, or otherwise referenced,
    # it may be garbage collected at any time.
    # This can lead to unexpected and inconsistent behavior.
    all_tasks.add(task)

    # To prevent keeping references to finished tasks forever,
    # make each task remove its own reference
    # from the set after completion.
    task.add_done_callback(all_tasks.discard)

    return task


class UniqueTask:
    """
    A class to manage spawning unique async tasks,
    ensuring only the latest one is running.
    """

    def __init__(self):
        self._task: Task[Any] | None = None

    def spawn(self, coro: Coroutine[None, None, Any]):
        """
        Spawns a new task, canceling the previous one if it exists.
        """
        self.cancel()
        self._task = create_task(coro)

    def cancel(self):
        """Cancels the previous task if it exists."""
        if self._task is not None and not self._task.done():
            self._task.cancel()

    def add_done_callback(self, callback: Callable[[Task[Any]], Any]):
        """
        Adds a callback to be called when the current task is done.
        """
        if self._task is not None:
            self._task.add_done_callback(callback)



import functools
from asyncio import get_event_loop
from collections.abc import Callable
from concurrent.futures import ProcessPoolExecutor
from multiprocessing import Manager, cpu_count
from multiprocessing.managers import SyncManager
from typing import ParamSpec, TypeVar

T = TypeVar("T")
P = ParamSpec("P")

PROCESS_COUNT = cpu_count()
sync_manager: SyncManager | None = None


def prepare_process_pool():
    global process_pool
    process_pool = ProcessPoolExecutor(PROCESS_COUNT)


def get_sync_manager() -> SyncManager:
    global sync_manager
    if sync_manager is None:
        sync_manager = Manager()
    return sync_manager


async def spawn_blocking(
    callable: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> T:
    """
    Executes the given callable in a separate process pool
    using `asyncio`'s `run_in_executor`.
    This function is intended for executing blocking or CPU-bound operations
    asynchronously outside `asyncio`'s event loop.

    Example:
    ```python
    # Define a blocking function
    def my_blocking_function(a, b):
        for _ in range(1000000):
            pass
        return a + b

    # Call the blocking function asynchronously using go
    result = await go(my_blocking_function, 10, 20)
    print(result)  # Output: 30
    ```
    """
    event_loop = get_event_loop()
    result = await event_loop.run_in_executor(
        process_pool,
        functools.partial(callable, *args, **kwargs),
    )
    return result



from .concurrency import UniqueTask, spawn
from .connect_event import outsource
from .info import PACKAGE_NAME, PACKAGE_PATH, PACKAGE_VERSION
from .parallelism import (
    PROCESS_COUNT,
    get_sync_manager,
    prepare_process_pool,
    spawn_blocking,
)

__all__ = [
    "PACKAGE_NAME",
    "PACKAGE_PATH",
    "PACKAGE_VERSION",
    "PROCESS_COUNT",
    "UniqueTask",
    "get_sync_manager",
    "outsource",
    "prepare_process_pool",
    "spawn",
    "spawn_blocking",
]



from PySide6.QtCore import QEvent
from PySide6.QtGui import QEnterEvent
from PySide6.QtWidgets import QPushButton, QWidget
from typing_extensions import override


class Gauge(QPushButton):
    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.is_mouse_over = False
        self.normal_text = ""
        self.lock_text = "🔒"

    @override
    def enterEvent(self, event: QEnterEvent):
        self.is_mouse_over = True
        self.setFlat(False)
        super().setText(self.lock_text)
        super().enterEvent(event)

    @override
    def leaveEvent(self, event: QEvent):
        self.is_mouse_over = False
        self.setFlat(True)
        super().setText(self.normal_text)
        super().leaveEvent(event)

    @override
    def setText(self, text: str):
        self.normal_text = text
        if self.is_mouse_over:
            super().setText(self.lock_text)
        else:
            super().setText(self.normal_text)
